# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, Optional, Tuple

import torch
from compressed_tensors.compressors.base import BaseCompressor
from compressed_tensors.compressors.quantized_compressors.base import (
    BaseQuantizationCompressor,
)
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import QuantizationArgs
from compressed_tensors.quantization.lifecycle.forward import dequantize, quantize
from torch import Tensor


__all__ = ["pack_fp4_to_uint8", "unpack_fp4_from_uint8", "NVFP4PackedCompressor"]

FLOAT_TO_E2M1 = [
    0.0,
    0.5,
    1.0,
    1.5,
    2.0,
    3.0,
    4.0,
    6.0,
]


@BaseCompressor.register(name=CompressionFormat.nvfp4_pack_quantized.value)
class NVFP4PackedCompressor(BaseQuantizationCompressor):
    """
    Implements compression of FP4 values. Weights of each quantized layer
    are packed into uint8. Only supports symmetric weight compression for now.
    """

    @property
    def compression_param_names(self) -> Tuple[str]:
        """
        Returns a tuple of compression parameter names introduced by
        the compressor during compression
        """
        return (
            "weight_packed",
            "weight_scale",
            "weight_zero_point",
            "weight_global_scale",
        )

    def compression_param_info(
        self,
        weight_shape: torch.Size,
        quantization_args: Optional[QuantizationArgs] = None,
    ) -> Dict[str, Tuple[torch.Size, torch.dtype]]:
        """
        Creates a dictionary of expected shapes and dtypes for each compression
            parameter used by the compressor

        :param weight_shape: uncompressed weight shape
        :param quantization_args: quantization parameters for the weight
        :return: dictionary mapping compressed parameter names to shape and dtype
        """
        output = {
            "weight_packed": (
                torch.Size((weight_shape[0], weight_shape[1] // 2)),
                torch.uint8,
            ),
        }
        return output

    def compress_scale(
        self,
        scale: Tensor,
        quantization_args: QuantizationArgs,
    ) -> Dict[str, torch.Tensor]:
        assert quantization_args.scale_dtype is not None
        return scale.to(quantization_args.scale_dtype)

    def compress_weight(
        self,
        weight: Tensor,
        scale: Tensor,
        global_scale: Tensor,
        quantization_args: QuantizationArgs,
        device: Optional[torch.device] = None,
        zero_point: Optional[torch.Tensor] = None,
        g_idx: Optional[torch.Tensor] = None,
    ) -> Dict[str, torch.Tensor]:
        quantized_weight = quantize(
            x=weight,
            scale=scale,
            global_scale=global_scale,
            zero_point=zero_point,
            args=quantization_args,
        )
        compressed_dict = {}
        weight_packed = pack_fp4_to_uint8(quantized_weight)
        if device is not None:
            weight_packed = weight_packed.to(device)
        compressed_dict["weight_packed"] = weight_packed
        compressed_dict["weight_scale"] = self.compress_scale(
            scale=scale, quantization_args=quantization_args
        )
        return compressed_dict

    def decompress_weight(
        self,
        compressed_data: Dict[str, Tensor],
        quantization_args: Optional[QuantizationArgs] = None,
    ) -> torch.Tensor:
        weight = compressed_data["weight_packed"]
        global_scale = compressed_data["weight_global_scale"]
        scale = compressed_data["weight_scale"]
        m, n = weight.shape
        # TODO: use a user provided dequant dtype
        unpacked = unpack_fp4_from_uint8(weight, m, n * 2)
        decompressed_weight = dequantize(
            x_q=unpacked, scale=scale, global_scale=global_scale, dtype=unpacked.dtype
        )

        return decompressed_weight


@BaseCompressor.register(name=CompressionFormat.mxfp4_pack_quantized.value)
class MXFP4PackedCompressor(NVFP4PackedCompressor):
    """
    Alias for mxfp4 quantized models
    """

    def compress_scale(
        self,
        scale: Tensor,
        quantization_args: QuantizationArgs,
    ) -> Dict[str, torch.Tensor]:
        assert quantization_args.scale_dtype is not None
        scale_exp = 127 + torch.floor(torch.log2(scale)).to(torch.int32) - 2
        return scale_exp.to(quantization_args.scale_dtype)

    def decompress_weight(
        self,
        compressed_data: Dict[str, Tensor],
        quantization_args: Optional[QuantizationArgs] = None,
    ) -> torch.Tensor:
        raise NotImplementedError("MXFP4 Decompression is currently not supported")


@torch.compile(fullgraph=True, dynamic=True)
def pack_fp4_to_uint8(x: torch.Tensor) -> torch.Tensor:
    """
    Packs a tensor with values in the fp4 range into uint8.
    As there are 16 valid fp4 values, two fp4 values can be
    packed into one uint8. Each fp4 value is mapped to its
    particular index (e.g. 0.5 is mapped to index 1, 6.0 is mapped
    to index 7) which is then represented using 4 bits. Consecutive
    pairs of 4 bits are then packed into an uint8.

    :param x: tensor to pack
    returns: a packed tensor in uint8
    """

    m, n = x.shape
    device = x.device

    if n % 2 != 0:
        raise ValueError(
            "tensor must have an even number of columns for nvfp4 compression"
        )

    # Create lookup table for FP4 values to indices
    # Map the absolute values to 0-7 indices
    kE2M1 = torch.tensor(FLOAT_TO_E2M1, device=device, dtype=x.dtype)

    # Find closest valid FP4 value index for each element
    abs_x = torch.abs(x)
    abs_diff_x = torch.abs(abs_x.unsqueeze(-1) - kE2M1)  # [m, n, 8]
    abs_indices = torch.argmin(abs_diff_x, dim=-1)  # [m, n]

    # Apply sign bit (bit 3) to get final 4-bit representation
    indices = abs_indices + (torch.signbit(x).to(torch.long) << 3)

    # Reshape to prepare for packing pairs of values
    indices = indices.reshape(-1)

    # Reshape to pair consecutive elements
    indices = indices.reshape(-1, 2)

    # Pack pairs of 4-bit values into 8-bit values
    packed = (indices[:, 0] | (indices[:, 1] << 4)).to(torch.uint8)

    return packed.reshape(m, n // 2)


kE2M1ToFloat = torch.tensor(
    [0.0, 0.5, 1.0, 1.5, 2.0, 3.0, 4.0, 6.0], dtype=torch.float32
)


# reference: : https://github.com/vllm-project/vllm/pull/16362
@torch.compile(fullgraph=True, dynamic=True)
def unpack_fp4_from_uint8(
    a: torch.Tensor, m: int, n: int, dtype: Optional[torch.dtype] = torch.bfloat16
) -> torch.Tensor:
    """
    Unpacks uint8 values into fp4. Each uint8 consists of two fp4 values
    (i.e. first four bits correspond to one fp4 value, last four correspond to a
    consecutive fp4 value). The bits represent an index, which are mapped to an fp4
    value.

    :param a: tensor to unpack
    :param m: original dim 0 size of the unpacked tensor
    :param n: original dim 1 size of the unpacked tensor
    :param dtype: dense dtype to cast the unpacked tensor to
    """
    assert a.dtype == torch.uint8

    # Vectorized nibble processing
    a_flat = a.flatten()
    high = (a_flat & 0xF0) >> 4  # Upper nibbles
    low = a_flat & 0x0F  # Lower nibbles

    # Combine nibbles for batch processing
    combined = torch.stack((low, high), dim=1).flatten()

    # Vectorized sign and magnitude extraction
    signs = (combined & 0x08).to(torch.bool)  # Sign bits
    abs_vals = (combined & 0x07).to(torch.long)  # Magnitude indices

    # Device-aware lookup and sign application
    kE2M1 = kE2M1ToFloat.to(device=a.device)
    values = kE2M1[abs_vals] * torch.where(signs, -1.0, 1.0)

    # Reshape to final form
    return values.reshape(m, n).to(dtype=dtype)



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch

import pytest
import torch.nn as nn

# Assuming the module is named "module_matching" - adjust import as needed
from compressed_tensors.utils import (
    InternalModule,
    get_lowest_common_ancestor_name,
    is_match,
    is_narrow_match,
    match_modules_set,
    match_named_modules,
    match_named_parameters,
)
from compressed_tensors.utils.match import _match_class, _match_name
from transformers import AutoModelForCausalLM


@pytest.fixture
def llama_stories_model():
    return AutoModelForCausalLM.from_pretrained(
        "Xenova/llama2.c-stories15M",
        torch_dtype="auto",
    )


class DummyModel(nn.Module):
    """Test model for unit tests. Weights are initialized on meta device"""

    def __init__(self):
        try:
            from accelerate import init_empty_weights
        except ImportError:
            pytest.skip("Skipping weight init requires accelerate")

        super().__init__()
        with init_empty_weights():
            self.layer1 = nn.Linear(10, 20)
            self.layer2 = nn.Linear(20, 30)
            self.norm = nn.LayerNorm(30)
            self.attention = nn.MultiheadAttention(30, 2)

            # Create nested structure
            self.transformer = nn.ModuleDict(
                {
                    "layers": nn.ModuleList(
                        [
                            nn.ModuleDict(
                                {
                                    "self_attn": nn.ModuleDict(
                                        {
                                            "q_proj": nn.Linear(30, 30),
                                            "k_proj": nn.Linear(30, 30),
                                            "v_proj": nn.Linear(30, 30),
                                        }
                                    ),
                                    "norm": nn.LayerNorm(30),
                                    "mlp": nn.Linear(30, 30),
                                }
                            )
                            for _ in range(3)
                        ]
                    )
                }
            )


class DummyMoEModel(nn.Module):
    """Test MoE model for unit tests. Weights are initialized on meta device"""

    def __init__(self, num_layers=2, num_experts=4):
        super().__init__()
        self.layers = nn.ModuleList(
            [
                nn.ModuleDict(
                    {
                        "post_attention_layernorm": nn.LayerNorm(3),
                        "mlp": nn.ModuleDict(
                            {
                                "experts": nn.ModuleList(
                                    [
                                        nn.ModuleDict(
                                            {
                                                "gate_proj": nn.Linear(3, 6),
                                                "up_proj": nn.Linear(3, 6),
                                                "down_proj": nn.Linear(6, 3),
                                            }
                                        )
                                        for _ in range(num_experts)
                                    ]
                                ),
                            }
                        ),
                    }
                )
                for _ in range(num_layers)
            ]
        )


class TestMatchName:
    """Test cases for _match_name function"""

    def test_exact_match(self):
        """Test exact string matching"""
        assert _match_name("layer1", "layer1")
        assert not _match_name("layer1", "layer2")
        assert _match_name(
            "transformer.layers.0.self_attn.q_proj",
            "transformer.layers.0.self_attn.q_proj",
        )

    def test_regex_match(self):
        """Test regex matching with "re:" prefix"""
        assert _match_name("layer1", "re:layer.*")
        assert _match_name("layer1", "re:^layer1$")
        assert not _match_name("layer1", "re:layer2")
        assert _match_name("transformer.layers.0.self_attn.q_proj", "re:.*q_proj")
        assert _match_name(
            "transformer.layers.0.self_attn.q_proj",
            "re:transformer\\.layers\\.\\d+\\.self_attn\\..*_proj$",
        )

    def test_empty_strings(self):
        """Test edge cases with empty strings"""
        assert _match_name("", "")
        assert not _match_name("layer1", "")
        assert not _match_name("", "layer1")

    def test_regex_special_characters(self):
        """Test regex with special characters"""
        assert _match_name("layer.1", "re:layer\\.1")
        assert _match_name("layer.1", "re:layer.1")  # . matches any char
        assert _match_name("layer_1", "re:layer_1")


class TestMatchClass:
    """Test cases for _match_class function"""

    def test_direct_class_match(self):
        """Test matching direct class names"""
        linear = nn.Linear(10, 20)
        assert _match_class(linear, "Linear")
        assert not _match_class(linear, "Conv2d")

        norm = nn.LayerNorm(10)
        assert _match_class(norm, "LayerNorm")
        assert not _match_class(norm, "BatchNorm1d")

    def test_parent_class_match(self):
        """Test matching parent class names"""
        linear = nn.Linear(10, 20)
        assert _match_class(linear, "Module")

        conv = nn.Conv2d(3, 16, 3)
        assert _match_class(conv, "Module")
        assert _match_class(conv, "_ConvNd")

    def test_non_torch_module(self):
        """Test with non-torch modules"""
        regular_object = object()
        assert not _match_class(regular_object, "object")  # not a torch.nn.Module

    def test_custom_module(self):
        """Test with custom module classes"""
        model = DummyModel()
        assert _match_class(model, "DummyModel")
        assert _match_class(model, "Module")

    def test_linear_base(self):
        """Test matching against vllm's LinearBase class"""

        class LinearBase(nn.Module):
            pass

        linear = LinearBase()
        assert _match_class(linear, "Linear")


class TestIsMatch:
    """Test cases for is_match function"""

    def test_name_match(self):
        """Test matching by name"""
        linear = nn.Linear(10, 20)
        assert is_match("layer1", linear, "layer1")
        assert not is_match("layer1", linear, "layer2")

    def test_class_match(self):
        """Test matching by class"""
        linear = nn.Linear(10, 20)
        assert is_match("layer1", linear, "Linear")
        assert not is_match("layer1", linear, "Conv2d")

    def test_combined_match(self):
        """Test that either name or class match works"""
        linear = nn.Linear(10, 20)
        assert is_match("layer1", linear, "layer1")  # name match
        assert is_match("layer1", linear, "Linear")  # class match
        assert not is_match("layer1", linear, "layer2")  # no match

    def test_regex_in_name_match(self):
        """Test regex matching in name"""
        linear = nn.Linear(10, 20)
        assert is_match("layer1", linear, "re:layer.*")
        assert not is_match("layer1", linear, "re:conv.*")

    def test_internal_module_match(self):
        """Test not matching internal modules"""

        class InternalLinear(InternalModule, nn.Linear):
            pass

        linear = InternalLinear(10, 20)
        assert not is_match("layer1", linear, "re:layer.*")

    def test_fused_mapping(self):
        """"""
        linear = nn.Linear(10, 20)
        mapping = {
            "qkv_proj": ["q_proj", "k_proj", "v_proj"],
            "gate_up_proj": ["gate_proj", "up_proj"],
        }

        assert is_match("dummy.qkv_proj", linear, "re:.*q_proj", fused=mapping)
        assert is_match("dummy.qkv_proj", linear, "re:.*k_proj", fused=mapping)
        assert is_match("dummy.qkv_proj", linear, "re:.*v_proj", fused=mapping)
        assert is_match("dummy.qkv_proj", linear, "Linear", fused=mapping)

        assert is_match("dummy.gate_up_proj", linear, "re:.*gate_proj", fused=mapping)
        assert is_match("dummy.gate_up_proj", linear, "re:.*up_proj", fused=mapping)
        assert is_match("dummy.gate_up_proj", linear, "Linear", fused=mapping)


class TestMatchNamedModules:
    """Test cases for match_named_modules function"""

    def test_exact_module_match(self):
        """Test matching modules by exact name"""
        model = DummyModel()
        matches = list(match_named_modules(model, ["layer1", "layer2"]))

        assert len(matches) == 2
        names = [name for name, _ in matches]
        assert "layer1" in names
        assert "layer2" in names

    def test_class_module_match(self):
        """Test matching modules by class name"""
        model = DummyModel()
        matches = list(match_named_modules(model, ["Linear"]))

        # Should find all Linear layers
        linear_modules = [
            module for _, module in matches if isinstance(module, nn.Linear)
        ]
        assert len(linear_modules) > 0

    def test_regex_module_match(self):
        """Test matching modules with regex patterns"""
        model = DummyModel()
        matches = list(match_named_modules(model, ["re:.*linear.*"]))

        # Should find layers with "linear" in name (case insensitive depends on model)
        assert len(matches) >= 0  # May be 0 if no "linear" in names

    def test_ignore_parameter(self):
        """Test ignoring specific modules"""
        model = DummyModel()
        matches_without_ignore = list(match_named_modules(model, ["Linear"]))
        matches_with_ignore = list(
            match_named_modules(model, ["Linear"], ignore=["layer1"])
        )

        # Should have fewer matches when ignoring layer1
        assert len(matches_with_ignore) <= len(matches_without_ignore)

        # layer1 should not be in ignored results
        ignored_names = [name for name, _ in matches_with_ignore]
        assert "layer1" not in ignored_names

    def test_empty_targets(self):
        """Test with empty targets list"""
        model = DummyModel()
        matches = list(match_named_modules(model, []))
        assert len(matches) == 0

    @patch("compressed_tensors.utils.match._LOGGER")
    def test_warn_on_fail(self, mock_logger):
        """Test warning when targets don"t match"""
        model = DummyModel()
        list(match_named_modules(model, ["nonexistent_module"], warn_on_fail=True))

        mock_logger.warning.assert_called_once()
        warning_msg = mock_logger.warning.call_args[0][0]
        assert "Could not match" in warning_msg
        assert "nonexistent_module" in warning_msg

    def test_internal_match(self):
        """Test not matching internal modules"""

        class InternalLinear(InternalModule, nn.Linear):
            pass

        linear = InternalLinear(10, 20)
        matches = list(match_named_modules(linear, ["re:.*"]))
        assert len(matches) == 0

    @pytest.mark.parametrize(
        "targets, ignore, expected_targets",
        [
            (
                ["re:model.layers.[01].self_attn.q_proj"],
                ["re:model.layers.1.self_attn.q_proj"],
                set(["model.layers.0.self_attn.q_proj"]),
            ),
            (
                ["re:model.layers.[01].self_attn.q_proj"],
                [],
                set(
                    [
                        "model.layers.0.self_attn.q_proj",
                        "model.layers.1.self_attn.q_proj",
                    ]
                ),
            ),
            (
                ["re:model.layers.[0-2].self_attn.q_proj"],
                ["re:model.layers.1.self_attn.q_proj"],
                set(
                    [
                        "model.layers.0.self_attn.q_proj",
                        "model.layers.2.self_attn.q_proj",
                    ]
                ),
            ),
            (
                ["model.layers.0.self_attn.q_proj"],
                ["model.layers.0.self_attn.q_proj"],
                set(),
            ),
            (
                ["re:model.layers.*.self_attn.q_proj"],
                ["re:model.layers.[01].self_attn.q_proj"],
                set(
                    f"model.layers.{layer_idx}.self_attn.q_proj"
                    for layer_idx in range(2, 6)
                ),
            ),
        ],
    )
    def test_expand_targets_with_llama_stories(
        self, llama_stories_model, targets, ignore, expected_targets
    ):
        expanded_targets = {
            name
            for name, _ in match_named_modules(llama_stories_model, targets, ignore)
        }
        assert expanded_targets == expected_targets


class TestMatchNamedParameters:
    """Test cases for match_named_parameters function"""

    def test_parameter_match(self):
        """Test matching parameters by name"""
        model = DummyModel()
        matches = list(match_named_parameters(model, ["layer1.weight", "layer1.bias"]))

        assert len(matches) == 2
        param_names = [name for name, _, _ in matches]
        assert "layer1.weight" in param_names
        assert "layer1.bias" in param_names

    def test_regex_parameter_match(self):
        """Test matching parameters with regex"""
        model = DummyModel()
        matches = list(match_named_parameters(model, ["re:.*weight$"]))

        # Should find all weight parameters
        weight_params = [name for name, _, _ in matches if name.endswith(".weight")]
        assert len(weight_params) > 0

    def test_ignore_parameters(self):
        """Test ignoring specific parameters"""
        model = DummyModel()
        matches_without_ignore = list(match_named_parameters(model, ["re:.*weight$"]))
        matches_with_ignore = list(
            match_named_parameters(model, ["re:.*weight$"], ignore=["layer1.weight"])
        )

        # Should have fewer matches when ignoring
        assert len(matches_with_ignore) < len(matches_without_ignore)

        # layer1.weight should not be in ignored results
        ignored_names = [name for name, _, _ in matches_with_ignore]
        assert "layer1.weight" not in ignored_names

    def test_parameter_return_values(self):
        """Test that function returns correct tuple values"""
        model = DummyModel()
        matches = list(match_named_parameters(model, ["layer1.weight"]))

        assert len(matches) == 1
        param_name, parent_module, param = matches[0]

        assert param_name == "layer1.weight"
        assert parent_module is model.layer1
        assert isinstance(param, nn.Parameter)
        assert param is model.layer1.weight

    @patch("compressed_tensors.utils.match._LOGGER")
    def test_warn_on_fail_parameters(self, mock_logger):
        """Test warning when parameter targets don"t match"""
        model = DummyModel()
        list(match_named_parameters(model, ["nonexistent.param"], warn_on_fail=True))

        mock_logger.warning.assert_called_once()
        warning_msg = mock_logger.warning.call_args[0][0]
        assert "Could not match" in warning_msg
        assert "nonexistent.param" in warning_msg

    def test_internal_match(self):
        """Test not matching internal modules"""

        class InternalLinear(InternalModule, nn.Linear):
            pass

        linear = InternalLinear(10, 20)
        matches = list(match_named_parameters(linear, ["re:.*"]))
        assert len(matches) == 0


class TestGetLowestCommonModuleName:
    """Test cases for get_lowest_common_ancestor_name function"""

    def test_multiple_modules(self):
        assert "abc" == get_lowest_common_ancestor_name(
            [
                "abc.a",
                "abc.b",
                "abc.c",
            ]
        )

    def test_single_module(self):
        assert "abc.abc" == get_lowest_common_ancestor_name(
            [
                "abc.abc",
            ]
        )

    def test_substring_modules(self):
        assert "abc" == get_lowest_common_ancestor_name(
            [
                "abc.abc",
                "abc.ab",
            ]
        )

    def test_parent_and_child_modules(self):
        assert "abc.abc" == get_lowest_common_ancestor_name(
            [
                "abc.abc.ab",
                "abc.abc",
            ]
        )

    def test_root(self):
        assert "" == get_lowest_common_ancestor_name(
            [
                "abc.abc",
                "b.abc",
            ]
        )

    def test_ignore_none(self):
        assert "abc.abc" == get_lowest_common_ancestor_name(
            [
                "abc.abc",
                None,
            ]
        )


class TestMatchModulesSet:
    """Test cases for match_modules_set function"""

    def test_simple_module_set(self):
        """Test matching simple module sets"""
        model = DummyModel()
        targets = [
            "re:.*self_attn.q_proj$",
            "re:.*self_attn.k_proj$",
            "re:.*self_attn.v_proj$",
        ]

        matches = list(match_modules_set(model, targets))

        # Should have 3 sets (one for each layer)
        assert len(matches) == 3

        # Each set should have 3 modules
        for module_set in matches:
            assert len(module_set) == 3
            assert all(isinstance(*m, nn.Linear) for m in module_set)

    def test_moe_module_match(self):
        """Test matching MoE modules with multiple experts per layer"""
        model = DummyMoEModel(num_layers=2, num_experts=4)

        # Test matching expert projections - each expert becomes its own set
        # because the parent context differs between experts
        targets = [
            "re:.*gate_proj$",
            "re:.*up_proj$",
        ]

        matches = list(match_modules_set(model, targets))

        # Should have 8 sets (2 layers * 4 experts)
        assert len(matches) == 8

        # Each set should have 2 target lists (gate_proj, up_proj)
        for expert_group in matches:
            assert len(expert_group) == 2
            gate_modules, up_modules = expert_group

            # Each target should have matched 1 module (single expert)
            assert len(gate_modules) == 1
            assert len(up_modules) == 1

            # All modules should be Linear layers
            assert isinstance(gate_modules[0], nn.Linear)
            assert isinstance(up_modules[0], nn.Linear)

    def test_moe_with_layernorm_match(self):
        """
        Test matching MoE modules with their corresponding layer norms.
        Including a layer-level module (layernorm) groups all experts in
        that layer together.
        """
        model = DummyMoEModel(num_layers=2, num_experts=3)

        # Match layer norm with expert projections - the layernorm is at layer level,
        # so it establishes a common parent context for all experts in that layer
        targets = [
            "re:.*post_attention_layernorm$",
            "re:.*gate_proj$",
            "re:.*up_proj$",
        ]

        matches = list(match_modules_set(model, targets))

        # Should have 2 layer groups (one per layer)
        assert len(matches) == 2

        for layer_group in matches:
            assert len(layer_group) == 3
            norm_modules, gate_modules, up_modules = layer_group

            # LayerNorm should have 1 module (single per layer)
            assert len(norm_modules) == 1
            assert isinstance(norm_modules[0], nn.LayerNorm)

            # Each projection should have 3 experts (all experts in the layer)
            assert len(gate_modules) == 3
            assert len(up_modules) == 3
            assert all(isinstance(m, nn.Linear) for m in gate_modules)
            assert all(isinstance(m, nn.Linear) for m in up_modules)

    def test_module_set_ordering(self):
        """Test that module sets maintain target ordering"""
        model = DummyModel()
        targets = [
            "re:.*v_proj$",  # v first
            "re:.*self_attn.q_proj$",  # q second
            "re:.*self_attn.k_proj$",
        ]  # k third

        matches = list(match_modules_set(model, targets))

        for module_set in matches:
            # Check that modules are returned in target order (v, q, k)
            v_proj, q_proj, k_proj = module_set
            v_proj, q_proj, k_proj = *v_proj, *q_proj, *k_proj
            # We can't easily check the exact modules, but can check they're all Linear
            assert all(isinstance(m, nn.Linear) for m in [v_proj, q_proj, k_proj])

    def test_incomplete_set_error(self):
        """Test error when unable to complete a set"""
        model = DummyModel()
        targets = ["layer1", "nonexistent_module"]

        with pytest.raises(
            ValueError, match="Found a final incomplete set with matches found for keys"
        ):
            list(match_modules_set(model, targets))

    def test_empty_targets_set(self):
        """Test with empty targets"""
        model = DummyModel()
        matches = list(match_modules_set(model, []))
        # Should yield one empty set for each module traversed?
        # Actually, with empty targets, we expect no matches
        assert len(matches) == 0

    def test_module_set_with_ignore(self):
        """Test module set matching with ignore parameter"""
        model = DummyModel()
        targets = ["re:.*self_attn.q_proj$", "re:.*self_attn.k_proj$"]
        ignore = ["re:transformer.layers.0.*"]  # Ignore first layer

        matches = list(match_modules_set(model, targets, ignore=ignore))

        # Should have 2 sets (layers 1 and 2, but not 0)
        assert len(matches) == 2

    def test_internal_match(self):
        """Test not matching internal modules"""

        class InternalLinear(InternalModule, nn.Linear):
            pass

        linear = InternalLinear(10, 20)
        matches = list(match_modules_set(linear, ["re:.*"]))
        assert len(matches) == 0


class TestIsNarrowMatch:
    def test_narrow_match_true_child_only(self):
        """
        Target matches the child module name but NOT its parent name.
        Should return True.
        """
        model = DummyModel()
        name = "transformer.layers.0.self_attn.q_proj"
        # Matches "...q_proj" but not "...self_attn"
        target = r"re:.*q_proj$"

        assert is_narrow_match(model, target, name)

    def test_narrow_match_false_when_parent_also_matches(self):
        """
        Target matches both the child and its parent name.
        Should return False because it's not a 'narrow' match.
        """
        model = DummyModel()
        name = "transformer.layers.0.self_attn.q_proj"
        # Broad target that also matches the parent "transformer.layers.0.self_attn"
        target = r"re:transformer\.layers\.0\..*"

        assert not is_narrow_match(model, target, name)

    def test_narrow_match_false_when_neither_matches(self):
        """
        Target matches neither the child nor the parent.
        Should return False.
        """
        model = DummyModel()
        name = "transformer.layers.0.self_attn.q_proj"
        target = r"re:this_does_not_exist$"

        assert not is_narrow_match(model, target, name)

    def test_narrow_match_iterable_targets_any_true(self):
        """
        With multiple targets: if any target narrowly matches the child,
        the function should return True.
        """
        model = DummyModel()
        name = "transformer.layers.0.self_attn.q_proj"
        # First target is broad (matches both child & parent -> narrow False),
        # second target is narrow (matches child only -> narrow True).
        targets = [
            r"re:transformer\.layers\.0\..*",
            r"re:.*q_proj$",
        ]

        assert is_narrow_match(model, targets, name)

    def test_narrow_match_with_explicit_module_argument(self):
        """
        Passing the module explicitly should behave the same as when it's
        retrieved from the model by name.
        """
        model = DummyModel()
        name = "transformer.layers.0.self_attn.q_proj"
        module = model.get_submodule(name)
        target = r"re:.*q_proj$"

        # Both ways should be True
        assert is_narrow_match(model, target, name)
        assert is_narrow_match(model, target, name, module=module)

    def test_narrow_match_top_level_behavior_documented(self):
        """
        (Behavior check) For a top-level module name without a dot, the current
        implementation derives parent_name == name, so parent==child.
        Then 'narrow' cannot be True because parent match mirrors child match.
        This test documents current behavior to guard against regressions.
        """
        model = DummyModel()
        name = "layer1"  # top-level module in DummyModel
        target = r"re:^layer1$"

        assert not is_narrow_match(model, target, name)


class TestIntegration:
    """Integration tests combining multiple functions"""

    def test_complex_model_matching(self):
        """Test matching on a more complex model structure"""
        model = DummyModel()

        # Test that we can find attention projection layers
        q_matches = list(match_named_modules(model, ["re:.*q_proj$"]))
        k_matches = list(match_named_modules(model, ["re:.*k_proj$"]))
        v_matches = list(match_named_modules(model, ["re:.*v_proj$"]))

        assert len(q_matches) == 3  # 3 layers
        assert len(k_matches) == 3
        assert len(v_matches) == 3

    def test_parameter_and_module_consistency(self):
        """Test that parameter and module matching are consistent"""
        model = DummyModel()

        # Get modules
        module_matches = list(match_named_modules(model, ["layer1"]))
        assert len(module_matches) == 1
        module_name, module = module_matches[0]

        # Get parameters from that module
        param_matches = list(match_named_parameters(model, [f"{module_name}.weight"]))
        assert len(param_matches) == 1
        param_name, parent_module, param = param_matches[0]

        # Check consistency
        assert parent_module is module
        assert param is module.weight

    def test_all_functions_with_regex(self):
        """Test all functions work with regex patterns"""
        model = DummyModel()
        regex_target = "re:.*Linear.*"

        # Should not crash and should handle regex consistently
        modules = list(match_named_modules(model, [regex_target]))
        params = list(match_named_parameters(model, [regex_target]))

        # Basic sanity checks
        assert isinstance(modules, list)
        assert isinstance(params, list)



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
import shutil
import tempfile
from collections import OrderedDict

import pytest
import torch
from compressed_tensors import PackedQuantizationCompressor
from compressed_tensors.compressors.quantized_compressors.pack_quantized import (
    pack_to_int32,
    unpack_from_int32,
)
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationConfig,
    QuantizationScheme,
    QuantizationStatus,
    QuantizationStrategy,
    apply_quantization_config,
)
from compressed_tensors.quantization.lifecycle.forward import fake_quantize
from compressed_tensors.quantization.quant_args import ActivationOrdering
from safetensors.torch import save_file
from torch.nn.modules import Linear, Sequential


def get_dummy_quant_config(
    num_bits=4, strategy=None, group_size=None, actorder=None, symmetric=True
) -> QuantizationConfig:
    config_groups = {
        "group_1": QuantizationScheme(
            targets=["Linear"],
            weights=QuantizationArgs(
                num_bits=num_bits,
                strategy=strategy,
                group_size=group_size,
                actorder=actorder,
                symmetric=symmetric,
            ),
        ),
    }
    return QuantizationConfig(config_groups=config_groups)


def make_dummy_g_idx(columns: int, group_size: int) -> torch.Tensor:
    perm = torch.randperm(columns)
    return torch.nn.Parameter(
        (torch.arange(columns, dtype=torch.int) // group_size)[perm],
        requires_grad=False,
    )


@pytest.mark.parametrize(
    "shape",
    [
        (512, 1024),
        (830, 545),
        (342, 512),
        (256, 700),
    ],
)
def test_quant_format(shape):
    dense_state_dict = {
        "dummy.weight": torch.rand(shape),
        "dummy.weight_scale": torch.tensor(0.01, dtype=torch.float32),
        "dummy.weight_zero_point": torch.tensor(0, dtype=torch.int8),
    }
    quant_config = get_dummy_quant_config()

    compressor = PackedQuantizationCompressor(config=quant_config)
    quantized_modules_to_scheme = {"dummy": quant_config.config_groups["group_1"]}
    compressed_state_dict = compressor.compress(
        dense_state_dict, names_to_scheme=quantized_modules_to_scheme
    )

    # compressed state_dict adds one entry for shape
    # but removes the zero points since we are symmetric
    assert len(dense_state_dict) == len(compressed_state_dict)

    # check compressed and packed
    assert compressed_state_dict["dummy.weight_packed"].dtype == torch.int32
    expected_rows = shape[0]
    expected_columns = math.ceil(shape[1] / 8)  # round each row up to nearest int32
    assert compressed_state_dict["dummy.weight_packed"].shape == (
        expected_rows,
        expected_columns,
    )

    assert torch.equal(compressed_state_dict["dummy.weight_shape"], torch.tensor(shape))
    assert compressed_state_dict["dummy.weight_scale"].dtype == torch.float32


@pytest.mark.parametrize(
    "value",
    [
        torch.tensor([[1, 2], [3, 4]]),
        torch.tensor([[1, 2, 3, 4, 5, 6, 7, 0], [-1, -2, -3, -4, -5, -6, -7, -8]]),
        (torch.rand((32, 100)) * 16 - 8),
    ],
)
def test_repack_4bit(value):
    value = value.to(torch.int8)
    shape = value.shape
    assert not torch.any(value > 7).item()
    assert not torch.any(value < -8).item()

    packed = pack_to_int32(value, 4)
    unpacked = unpack_from_int32(packed, 4, shape)
    assert torch.equal(value, unpacked)


@pytest.mark.parametrize(
    "value",
    [
        torch.tensor([[30, 40], [50, 60]]),
        torch.tensor(
            [[10, 15, 20, 25, 30, 35, 40, 45], [-10, -20, -30, -40, -50, -60, -70, -80]]
        ),
        (torch.rand((32, 100)) * 256 - 128),
    ],
)
def test_repack_8bit(value):
    value = value.to(torch.int8)
    shape = value.shape
    assert not torch.any(value > 127).item()
    assert not torch.any(value < -128).item()

    packed = pack_to_int32(value, 8)
    unpacked = unpack_from_int32(packed, 8, shape)
    assert torch.equal(value, unpacked)


@pytest.mark.parametrize("num_bits", [4, 8])
def test_reload_match(tmp_path, num_bits):
    dense_state_dict = {
        "dummy.weight": torch.rand((511, 350)),
        "dummy.weight_scale": torch.tensor(0.01, dtype=torch.float32),
        "dummy.weight_zero_point": torch.tensor(0, dtype=torch.int8),
        "dummy2.weight": torch.rand((128, 280)),
        "dummy2.weight_scale": torch.tensor(0.02, dtype=torch.float32),
        "dummy2.weight_zero_point": torch.tensor(15, dtype=torch.int8),
    }

    # pack-compressor only needs the number of bits from the quant-args to decompress
    # all other information is extracted from the compressed data directly
    quant_config = get_dummy_quant_config(num_bits, symmetric=False)

    compressor = PackedQuantizationCompressor(config=quant_config)
    quantized_modules_to_scheme = {
        "dummy": quant_config.config_groups["group_1"],
        "dummy2": quant_config.config_groups["group_1"],
    }

    compressed_state_dict = compressor.compress(
        dense_state_dict.copy(), names_to_scheme=quantized_modules_to_scheme
    )
    save_file(compressed_state_dict, tmp_path / "model.safetensors")

    reconstructed_dense = {}
    with tempfile.TemporaryDirectory():
        reconstructed_dense_gen = compressor.decompress(
            tmp_path, names_to_scheme=quantized_modules_to_scheme
        )
        for name, value in reconstructed_dense_gen:
            reconstructed_dense[name] = value

    fake_quant_dummy = fake_quantize(
        dense_state_dict["dummy.weight"],
        scale=dense_state_dict["dummy.weight_scale"],
        zero_point=dense_state_dict["dummy.weight_zero_point"],
        args=quantized_modules_to_scheme["dummy"].weights,
    )
    assert torch.equal(
        fake_quant_dummy, reconstructed_dense["dummy"].get("weight").to(torch.float32)
    )

    fake_quant_dummy2 = fake_quantize(
        dense_state_dict["dummy2.weight"],
        scale=dense_state_dict["dummy2.weight_scale"],
        zero_point=dense_state_dict["dummy2.weight_zero_point"],
        args=quantized_modules_to_scheme["dummy2"].weights,
    )
    assert torch.equal(
        fake_quant_dummy2, reconstructed_dense["dummy2"].get("weight").to(torch.float32)
    )

    shutil.rmtree(tmp_path)


@pytest.mark.parametrize(
    "strategy",
    {QuantizationStrategy.GROUP, QuantizationStrategy.CHANNEL},
)
def test_asymmetric_packed_support(strategy):
    shape = (1024, 1024)

    group_size = None
    if strategy == QuantizationStrategy.GROUP:
        group_size = 128

    if strategy == QuantizationStrategy.CHANNEL:
        expected_shape = (shape[0], 1)
    elif strategy == QuantizationStrategy.GROUP:
        num_groups = shape[1] // group_size
        expected_shape = (shape[0], max(num_groups, 1))

    dense_state_dict = {
        "dummy.weight": torch.rand(shape),
        "dummy.weight_scale": torch.rand(expected_shape).to(torch.float32),
        "dummy.weight_zero_point": torch.rand(expected_shape).to(torch.int8),
    }

    quant_config = get_dummy_quant_config(
        strategy=strategy.value, symmetric=False, group_size=group_size
    )

    compressor = PackedQuantizationCompressor(config=quant_config)
    quantized_modules_to_scheme = {"dummy": quant_config.config_groups["group_1"]}
    compressed_state_dict = compressor.compress(
        dense_state_dict, names_to_scheme=quantized_modules_to_scheme
    )

    # compressed state_dict adds one entry for shape
    assert len(dense_state_dict) + 1 == len(compressed_state_dict)
    assert compressed_state_dict["dummy.weight_packed"].dtype == torch.int32
    assert compressed_state_dict["dummy.weight_zero_point"].dtype == torch.int32
    assert compressed_state_dict["dummy.weight_scale"].dtype == torch.float32

    # check weight compressed and packed
    expected_rows = shape[0]
    expected_columns = math.ceil(shape[1] / 8)  # round each row up to nearest int32
    assert compressed_state_dict["dummy.weight_packed"].shape == (
        expected_rows,
        expected_columns,
    )
    assert torch.equal(compressed_state_dict["dummy.weight_shape"], torch.tensor(shape))

    # check zp compressed and packed
    packed_size_zp = math.ceil(shape[0] / 8)
    zp_factor = group_size if strategy == QuantizationStrategy.GROUP else shape[-1]
    assert compressed_state_dict["dummy.weight_zero_point"].shape == (
        packed_size_zp,
        shape[-1] // zp_factor,
    )


@pytest.mark.parametrize(
    "actorder",
    [
        ActivationOrdering.GROUP,
        ActivationOrdering.WEIGHT,
        None,
    ],
)
def test_actorder_reload_match(actorder, tmp_path, mock_per_group_calibration):
    model = Sequential(OrderedDict([("dummy", Linear(512, 1024, bias=None))]))
    group_size = 128
    quant_config = get_dummy_quant_config(
        strategy="group", group_size=group_size, actorder=actorder
    )
    apply_quantization_config(model, quant_config)

    # run calibration
    model.quantization_status = QuantizationStatus.CALIBRATION
    mock_per_group_calibration(
        model.dummy, base_name="weight", value=model.dummy.weight, group_size=group_size
    )
    # apply gptq
    if actorder == ActivationOrdering.GROUP:
        init_g_idx = make_dummy_g_idx(512, group_size)
        model.dummy.register_parameter("weight_g_idx", init_g_idx)

    # compress
    compressor = PackedQuantizationCompressor(config=quant_config)
    quantized_modules_to_scheme = {
        "dummy": quant_config.config_groups["group_1"],
    }
    compressed_state_dict = compressor.compress(
        model.state_dict(), names_to_scheme=quantized_modules_to_scheme
    )
    save_file(compressed_state_dict, tmp_path / "model.safetensors")

    # decompress
    reconstructed_dense_gen = compressor.decompress(
        tmp_path, names_to_scheme=quantized_modules_to_scheme
    )
    reconstructed_dense = {}
    for name, value in reconstructed_dense_gen:
        reconstructed_dense[name] = value

    fake_quant_dummy = fake_quantize(
        model.dummy.weight,
        scale=model.dummy.weight_scale,
        zero_point=model.dummy.weight_zero_point,
        g_idx=getattr(model.dummy, "weight_g_idx", None),
        args=quantized_modules_to_scheme["dummy"].weights,
    )
    assert torch.equal(fake_quant_dummy, reconstructed_dense["dummy"].get("weight"))

    shutil.rmtree(tmp_path)


@pytest.mark.parametrize(
    "num_bits,values,expected_values",
    [
        (
            4,
            torch.tensor([[1]]),
            torch.tensor([[9]], dtype=torch.int32),
        ),
        (
            8,
            torch.tensor([[1]]),
            torch.tensor([[129]], dtype=torch.int32),
        ),
        # 0000 0000 0000 0000 1100 1011 1010 1001
        (4, torch.tensor([[1, 2, 3, 4]]), torch.tensor([[52137]], dtype=torch.int32)),
        # 0111 0110 0101 0100 0011 0010 0001 0000
        (
            4,
            torch.tensor([[-8, -7, -6, -5, -4, -3, -2, -1]]),
            torch.tensor([[1985229328]], dtype=torch.int32),
        ),
        # 10000100 10000011 10000010 10000001
        (
            8,
            torch.tensor([[1, 2, 3, 4]]),
            torch.tensor([[-2071756159]], dtype=torch.int32),
        ),
        # 00000011 00000010 00000001 00000000
        (
            8,
            torch.tensor([[-128, -127, -126, -125]]),
            torch.tensor([[50462976]], dtype=torch.int32),
        ),
        (
            4,
            torch.tensor([[-8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4]]),
            torch.tensor([[1985229328, 52137]], dtype=torch.int32),
        ),
        (
            4,
            torch.tensor(
                [
                    [-8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, -8, -8, -8, -8],
                    [1, 2, 3, 4, -8, -8, -8, -8, -8, -7, -6, -5, -4, -3, -2, -1],
                ]
            ),
            torch.tensor([[1985229328, 52137], [52137, 1985229328]], dtype=torch.int32),
        ),
        (
            8,
            torch.tensor(
                [
                    [1, 2, 3, 4],
                    [-128, -127, -126, -125],
                ]
            ),
            torch.tensor([[-2071756159], [50462976]], dtype=torch.int32),
        ),
        (
            8,
            torch.tensor(
                [
                    [1, 2, 3, 4, -128, -127, -126, -125],
                    [-128, -127, -126, -125, 1, 2, 3, 4],
                ]
            ),
            torch.tensor(
                [[-2071756159, 50462976], [50462976, -2071756159]], dtype=torch.int32
            ),
        ),
    ],
)
def test_pack_to_int32(num_bits, values, expected_values):
    values = values.to(torch.int8)
    packed_values = pack_to_int32(values, num_bits)
    assert torch.equal(packed_values, expected_values)
    assert packed_values.dtype == expected_values.dtype


@pytest.mark.parametrize(
    "num_bits,values,expected_tensor",
    [
        (
            4,
            torch.tensor([[9]], dtype=torch.int32),
            torch.tensor([[1]], dtype=torch.int8),
        ),
        (
            8,
            torch.tensor([[129]], dtype=torch.int32),
            torch.tensor([[1]], dtype=torch.int8),
        ),
        (
            4,
            torch.tensor([[52137]], dtype=torch.int32),
            torch.tensor([[1, 2, 3, 4]], dtype=torch.int8),
        ),
        (
            4,
            torch.tensor([[1985229328]], dtype=torch.int32),
            torch.tensor([[-8, -7, -6, -5, -4, -3, -2, -1]], dtype=torch.int8),
        ),
        (
            8,
            torch.tensor([[-2071756159]], dtype=torch.int32),
            torch.tensor([[1, 2, 3, 4]], dtype=torch.int8),
        ),
        (
            8,
            torch.tensor([[50462976]], dtype=torch.int32),
            torch.tensor([[-128, -127, -126, -125]], dtype=torch.int8),
        ),
        (
            4,
            torch.tensor([[1985229328, 52137]], dtype=torch.int32),
            torch.tensor(
                [[-8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4]], dtype=torch.int8
            ),
        ),
        (
            4,
            torch.tensor([[1985229328, 52137], [52137, 1985229328]], dtype=torch.int32),
            torch.tensor(
                [
                    [-8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, -8, -8, -8, -8],
                    [1, 2, 3, 4, -8, -8, -8, -8, -8, -7, -6, -5, -4, -3, -2, -1],
                ],
                dtype=torch.int8,
            ),
        ),
        (
            8,
            torch.tensor([[-2071756159], [50462976]], dtype=torch.int32),
            torch.tensor(
                [
                    [1, 2, 3, 4],
                    [-128, -127, -126, -125],
                ],
                dtype=torch.int8,
            ),
        ),
        (
            8,
            torch.tensor(
                [[-2071756159, 50462976], [50462976, -2071756159]], dtype=torch.int32
            ),
            torch.tensor(
                [
                    [1, 2, 3, 4, -128, -127, -126, -125],
                    [-128, -127, -126, -125, 1, 2, 3, 4],
                ],
                dtype=torch.int8,
            ),
        ),
    ],
)
def test_unpack_from_int32(num_bits, values, expected_tensor):
    unpacked_tensor = unpack_from_int32(values, num_bits, expected_tensor.shape)
    assert torch.equal(unpacked_tensor, unpacked_tensor)
    assert unpacked_tensor.dtype == unpacked_tensor.dtype


@pytest.mark.parametrize(
    "strategy,group_size",
    [
        (QuantizationStrategy.GROUP, 128),
        (QuantizationStrategy.CHANNEL, None),
    ],
)
def test_asymmetric_zero_point_decompression(strategy, group_size, tmp_path):
    """
    Test that zero-point packing and unpacking works correctly for asymmetric
    quantization with GROUP and CHANNEL strategies.
    """
    shape = (512, 1024)

    if strategy == QuantizationStrategy.CHANNEL:
        expected_zp_shape = (shape[0], 1)
    elif strategy == QuantizationStrategy.GROUP:
        num_groups = shape[1] // group_size
        expected_zp_shape = (shape[0], max(num_groups, 1))

    dense_state_dict = {
        "dummy.weight": torch.randn(shape),
        "dummy.weight_scale": torch.rand(expected_zp_shape).to(torch.float32),
        "dummy.weight_zero_point": torch.randint(-8, 8, expected_zp_shape).to(
            torch.int8
        ),
    }

    quant_config = get_dummy_quant_config(
        num_bits=4, strategy=strategy.value, symmetric=False, group_size=group_size
    )

    compressor = PackedQuantizationCompressor(config=quant_config)
    quantized_modules_to_scheme = {"dummy": quant_config.config_groups["group_1"]}
    compressed_state_dict = compressor.compress(
        dense_state_dict.copy(), names_to_scheme=quantized_modules_to_scheme
    )

    assert "dummy.weight_zero_point" in compressed_state_dict
    assert compressed_state_dict["dummy.weight_zero_point"].dtype == torch.int32

    save_file(compressed_state_dict, tmp_path / "model.safetensors")

    reconstructed_dense_gen = compressor.decompress(
        tmp_path, names_to_scheme=quantized_modules_to_scheme
    )
    reconstructed_dense = {}
    for name, value in reconstructed_dense_gen:
        reconstructed_dense[name] = value

    assert "dummy" in reconstructed_dense
    assert "weight" in reconstructed_dense["dummy"]

    assert reconstructed_dense["dummy"]["weight"].shape == shape

    shutil.rmtree(tmp_path)


@pytest.mark.parametrize(
    "num_bits,strategy",
    [
        (4, QuantizationStrategy.GROUP),
        (4, QuantizationStrategy.CHANNEL),
        (8, QuantizationStrategy.GROUP),
        (8, QuantizationStrategy.CHANNEL),
    ],
)
def test_zero_point_pack_unpack_consistency(num_bits, strategy):
    """
    Test that packing and unpacking zero-points preserves values correctly.
    """
    if strategy == QuantizationStrategy.GROUP:
        shape = (512, 8)
    else:
        shape = (512, 1)

    max_val = (1 << (num_bits - 1)) - 1
    min_val = -(1 << (num_bits - 1))
    original_zp = torch.randint(min_val, max_val + 1, shape).to(torch.int8)

    packed_zp = pack_to_int32(original_zp, num_bits, packed_dim=0)

    unpacked_zp = unpack_from_int32(packed_zp, num_bits, shape, packed_dim=0)

    assert torch.equal(original_zp, unpacked_zp)
    assert unpacked_zp.dtype == torch.int8



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Callable, Optional

from compressed_tensors.modeling.kvcache import initialize_hooked_kv_cache
from compressed_tensors.quantization.lifecycle.forward import forward_quantize
from compressed_tensors.utils import getattr_chain
from compressed_tensors.utils.internal import InternalModule
from torch import Tensor
from torch.nn import Module
from torch.utils.hooks import RemovableHandle
from transformers import PretrainedConfig, PreTrainedModel
from transformers.masking_utils import ALL_MASK_ATTENTION_FUNCTIONS
from transformers.modeling_utils import ALL_ATTENTION_FUNCTIONS


__all__ = [
    "QuantizedAttentionImpl",
    "initialize_hooked_attention",
    "register_query_hook",
    "IMPL_ATTR",
]


IMPL_ATTR = "impl"
HOOKED_ATTENTION_NAME = "ct_hooked_attention"


class QuantizedAttentionImpl(InternalModule):
    """
    QuantizedAttentionImpl module which wraps the functionality of the original
    attention implementation. Unlike the original attention function, this
    implementation is a `torch.nn.Module` which can be hooked to trigger
    transforms and calibration hooks.

    This module works by being registered as a submodule to attention modules via
    `initialize_hooked_attention`, registering a new attention implementation function
    which calls this module, then setting the model attention implementation to the new
    function. After triggering hooks and quantization, this module calls the original
    attention implementation function.
    """

    _original_impl = "eager"

    def __init__(self, config: PretrainedConfig):
        super().__init__()
        self.config = config

    def forward(
        self,
        module: Module,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        *args,
        **kwargs,
    ):
        # quantization
        quant_args_attr = "quantization_scheme.input_activations"
        quant_args = getattr_chain(module, quant_args_attr, None)
        quant_enabled = getattr(module, "quantization_enabled", True)
        if quant_args is not None and quant_enabled:
            query = forward_quantize(module, query, "q", quant_args)

        # original attention
        return ALL_ATTENTION_FUNCTIONS[QuantizedAttentionImpl._original_impl](
            module,
            query,
            key,
            value,
            *args,
            **kwargs,
        )


# ----- initialize ----- #


def _hooked_attention(module: Module, *args, **kwargs):
    assert hasattr(module, IMPL_ATTR), (
        f"Using {HOOKED_ATTENTION_NAME} attention implementation, "
        f"but attention module does not have {IMPL_ATTR} submodule."
    )

    return getattr(module, IMPL_ATTR)(module, *args, **kwargs)


def initialize_hooked_attention(model: PreTrainedModel, module: Module):
    """
    Initialize `QuantizedAttentionImpl` and `QuantizedKVCache` instances
    attached to attention. Assumes that only one model is hooked at a time.

    :param model: parent model of attention module
    :param module: attention module to initialize with
    """
    if not hasattr(module, IMPL_ATTR):
        module.register_module(IMPL_ATTR, QuantizedAttentionImpl(model.config))

    if model.config._attn_implementation != HOOKED_ATTENTION_NAME:
        QuantizedAttentionImpl._original_impl = model.config._attn_implementation
        original_mask = ALL_MASK_ATTENTION_FUNCTIONS[model.config._attn_implementation]

        ALL_ATTENTION_FUNCTIONS.register(HOOKED_ATTENTION_NAME, _hooked_attention)
        ALL_MASK_ATTENTION_FUNCTIONS.register(HOOKED_ATTENTION_NAME, original_mask)
        model.set_attn_implementation(HOOKED_ATTENTION_NAME)
        assert model.config._attn_implementation == HOOKED_ATTENTION_NAME

    initialize_hooked_kv_cache(model, module)


# ----- hooks ----- #


def register_query_hook(
    module: Module, hook: Callable[[Module, Tensor], Optional[Tensor]]
) -> RemovableHandle:
    """
    Register a hook which takes post-rope query states as an argument and
    returns the modified query states or `None`

    :param module: attention module to add hook to
    :param hook: query hook function
    """
    impl: QuantizedAttentionImpl = getattr(module, IMPL_ATTR)

    def _hook(impl: QuantizedAttentionImpl, args, kwargs):
        bound = inspect.signature(impl.forward).bind(*args, **kwargs)
        value = hook(module, bound.arguments["query"])
        if value is not None:
            bound.arguments["query"] = value

        return bound.args, bound.kwargs

    return impl.register_forward_pre_hook(_hook, with_kwargs=True)



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import ceil

import pytest
import torch
from compressed_tensors.quantization.utils import calculate_qparams
from compressed_tensors.utils.offload import update_parameter_data


def _get_dim(dim: int, value: torch.Tensor):
    if isinstance(dim, int):
        dim = [dim]
        dim = set(dim)

    reduce_dims = tuple(idx for idx in range(value.ndim) if idx not in dim)
    return reduce_dims


@pytest.fixture
def mock_per_group_calibration():
    def update_scale_zp(
        module: torch.nn.Module, base_name: str, value: torch.Tensor, group_size: int
    ):
        quantization_scheme = getattr(module, "quantization_scheme", None)
        if not quantization_scheme:
            # no quantization scheme nothing to do
            return

        arg_name = "weights" if base_name == "weight" else f"{base_name}_activations"
        args = getattr(quantization_scheme, arg_name, None)

        rows = value.shape[0]
        columns = value.shape[1]
        num_groups = int(ceil(columns / group_size))

        scale = torch.zeros((rows, num_groups), dtype=value.dtype, device=value.device)
        zp_dtype = args.pytorch_dtype()
        zp = torch.zeros((rows, num_groups), dtype=zp_dtype, device=value.device)

        group_sizes = torch.full((num_groups,), group_size, dtype=torch.int)
        end = 0
        for group_index, group_count in enumerate(group_sizes):
            start = end
            end = start + group_count
            dim = _get_dim(
                0,
                value[:, start:end],
            )
            min_val = torch.amin(value, dim=dim, keepdims=True)
            max_val = torch.amax(value, dim=dim, keepdims=True)
            scale_out, zp_out = calculate_qparams(min_val, max_val, args)

            scale[:, group_index] = scale_out.squeeze(1)
            zp[:, group_index] = zp_out.squeeze(1)

        update_parameter_data(module, scale, f"{base_name}_scale")
        update_parameter_data(module, zp, f"{base_name}_zero_point")

    return update_scale_zp


@pytest.fixture
def mock_per_channel_calibration():
    def update_scale_zp(module: torch.nn.Module, base_name: str, value: torch.Tensor):
        quantization_scheme = getattr(module, "quantization_scheme", None)
        if not quantization_scheme:
            # no quantization scheme nothing to do
            return

        arg_name = "weights" if base_name == "weight" else f"{base_name}_activations"

        args = getattr(quantization_scheme, arg_name, None)
        dim = _get_dim(0, value)
        min_val = torch.amin(value, dim=dim, keepdims=True)
        max_val = torch.amax(value, dim=dim, keepdims=True)
        scale, zp = calculate_qparams(min_val, max_val, args)
        update_parameter_data(module, scale, f"{base_name}_scale")
        update_parameter_data(module, zp, f"{base_name}_zero_point")

    return update_scale_zp


@pytest.fixture
def mock_per_tensor_calibration():
    def update_scale_zp(module: torch.nn.Module, base_name: str, value: torch.Tensor):
        quantization_scheme = getattr(module, "quantization_scheme", None)
        if not quantization_scheme:
            # no quantization scheme nothing to do
            return

        arg_name = "weights" if base_name == "weight" else f"{base_name}_activations"
        args = getattr(quantization_scheme, arg_name, None)

        # per tensor quantization just calls calculate_qparams directly
        min_val, max_val = torch.aminmax(value)
        scale, zp = calculate_qparams(min_val, max_val, args)
        update_parameter_data(module, scale, f"{base_name}_scale")
        update_parameter_data(module, zp, f"{base_name}_zero_point")

    return update_scale_zp



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch
from compressed_tensors.quantization import (
    FP8_E4M3_DATA,
    QuantizationScheme,
    forward_quantize,
    initialize_module_for_quantization,
    initialize_qparams,
)
from compressed_tensors.quantization.quant_args import QuantizationArgs
from compressed_tensors.quantization.quant_config import QuantizationStatus
from tests.mock_observer import MockMinMaxObserver


@pytest.mark.parametrize(
    "args,exp_min_val,exp_max_val,exp_quant,exp_loss",
    [
        (
            QuantizationArgs(
                num_bits=4,
                type="int",
                symmetric=True,
                strategy="tensor",  # equivalent to token
            ),
            torch.tensor([0.0]),
            torch.tensor([23.0]),
            torch.tensor(
                [
                    [0.0000, 0.0000, 3.0625, 3.0625, 3.0625, 6.1250],
                    [6.1250, 6.1250, 9.1875, 9.1875, 9.1875, 12.2500],
                    [12.2500, 12.2500, 15.3125, 15.3125, 15.3125, 18.3750],
                    [18.3750, 18.3750, 21.5000, 21.5000, 21.5000, 21.5000],
                ],
                dtype=torch.bfloat16,
            ),
            0.85,
        ),
        # token is not supported
        (
            QuantizationArgs(
                num_bits=4,
                type="int",
                symmetric=True,
                strategy="channel",
            ),
            torch.tensor([[0], [6], [12], [18]]),
            torch.tensor([[5], [11], [17], [23]]),
            torch.tensor(
                [
                    [0.0000, 1.3359, 2.0000, 2.6719, 4.0000, 4.6875],
                    [5.8750, 7.3438, 7.3438, 8.8125, 10.2500, 10.2500],
                    [11.3125, 13.6250, 13.6250, 15.8750, 15.8750, 15.8750],
                    [18.3750, 18.3750, 21.5000, 21.5000, 21.5000, 21.5000],
                ],
                dtype=torch.bfloat16,
            ),
            0.45,
        ),
        (
            QuantizationArgs(
                num_bits=4,
                type="int",
                symmetric=True,
                strategy="group",
                group_size=3,
            ),
            torch.tensor([[0, 3], [6, 9], [12, 15], [18, 21]]),
            torch.tensor([[2, 5], [8, 11], [14, 17], [20, 23]]),
            torch.tensor(
                [
                    [0.0000, 1.0703, 1.8750, 2.6719, 4.0000, 4.6875],
                    [6.4375, 7.5000, 7.5000, 8.8125, 10.2500, 10.2500],
                    [11.1875, 13.0625, 13.0625, 15.8750, 15.8750, 15.8750],
                    [18.7500, 18.7500, 18.7500, 21.5000, 21.5000, 21.5000],
                ],
            ),
            0.45,
        ),
        (
            QuantizationArgs(
                num_bits=4,
                type="float",  # tensor group requires FP4
                symmetric=True,
                strategy="tensor_group",  # requires float4
                group_size=3,
                scale_dtype=FP8_E4M3_DATA.dtype,
                zp_dtype=FP8_E4M3_DATA.dtype,
            ),
            torch.tensor([[0, 3], [6, 9], [12, 15], [18, 21]]),
            torch.tensor([[2, 5], [8, 11], [14, 17], [20, 23]]),
            torch.tensor(
                [
                    [0.0000, 1.0234, 2.0469, 3.2812, 3.2812, 4.9375],
                    [5.4688, 8.1875, 8.1875, 10.6875, 10.6875, 10.6875],
                    [9.8750, 14.7500, 14.7500, 16.3750, 16.3750, 16.3750],
                    [19.7500, 19.7500, 19.7500, 23.0000, 23.0000, 23.0000],
                ],
            ),
            1.1,
        ),
        (
            QuantizationArgs(
                num_bits=4,
                type="int",
                symmetric=True,
                strategy="block",
                block_structure=[2, 3],
            ),
            torch.tensor([[0, 3], [12, 15]]),
            torch.tensor([[8, 11], [20, 23]]),
            torch.tensor(
                [
                    [0.0000, 1.0703, 2.1406, 2.9375, 4.4062, 4.4062],
                    [6.4375, 7.5000, 7.5000, 8.8125, 10.2500, 10.2500],
                    [10.6875, 13.3750, 13.3750, 15.3125, 15.3125, 18.3750],
                    [18.7500, 18.7500, 18.7500, 21.5000, 21.5000, 21.5000],
                ],
            ),
            0.5,
        ),
    ],
)
def test_static_weight_quantization(
    args, exp_min_val, exp_max_val, exp_quant, exp_loss
):
    """
    weight = tensor([[ 0,  1,  2,  3,  4,  5],
                     [ 6,  7,  8,  9, 10, 11],
                     [12, 13, 14, 15, 16, 17],
                     [18, 19, 20, 21, 22, 23]])
    """
    # set up weight
    input_size, output_size = 6, 4
    linear = torch.nn.Linear(input_size, output_size, bias=False)
    linear.weight.data = torch.arange(
        input_size * output_size, dtype=torch.bfloat16
    ).reshape(output_size, input_size)

    # initialize quantization parameters
    scheme = QuantizationScheme(targets=[], weights=args)
    initialize_module_for_quantization(linear, scheme)
    assert getattr(linear, "quantization_scheme") is scheme
    linear.weight_observer = MockMinMaxObserver("weight", args, linear)

    # calibrate_global_scale
    if hasattr(linear, "weight_global_scale"):
        global_scale = linear.weight_observer.get_global_scale(linear.weight)
        linear.weight_global_scale.data = global_scale

    # calibrate quantization parameters
    scale, zero_point = linear.weight_observer(linear.weight)
    linear.weight_scale.data = scale
    linear.weight_zero_point.data = zero_point
    assert torch.equal(linear.weight_observer.min_vals, exp_min_val)
    assert torch.equal(linear.weight_observer.max_vals, exp_max_val)

    # forward pass
    input = torch.eye(input_size, dtype=torch.bfloat16)
    output = linear(input)

    assert torch.allclose(output.T, exp_quant.to(output.dtype))
    assert torch.nn.functional.mse_loss(output.T, linear.weight) <= exp_loss


@pytest.mark.parametrize(
    "args,exp_min_val,exp_max_val,exp_quant,exp_loss",
    [
        (
            QuantizationArgs(
                num_bits=4,
                type="int",
                symmetric=True,
                strategy="tensor",
            ),
            torch.tensor([0.0]),
            torch.tensor([11.0]),
            torch.tensor(
                [
                    [
                        [0.0000, 1.4688, 1.4688, 2.9375, 4.4062, 4.4062],
                        [5.8750, 7.3438, 7.3438, 8.8125, 10.2500, 10.2500],
                    ]
                ]
            ),
            0.2,
        ),
        # static token is not supported
        # channel is not supported
        # group is not supported
        (
            QuantizationArgs(
                num_bits=4,
                type="float",  # must be fp4
                symmetric=True,
                strategy="tensor_group",
                dynamic="local",
                group_size=3,
                scale_dtype=FP8_E4M3_DATA.dtype,
                zp_dtype=FP8_E4M3_DATA.dtype,
            ),
            None,
            None,
            torch.tensor(
                [
                    [
                        [0.0000, 0.9844, 1.9688, 3.4062, 3.4062, 5.1250],
                        [5.2500, 7.8750, 7.8750, 7.3438, 11.0000, 11.0000],
                    ]
                ]
            ),
            0.5,
        ),
        # block is not supported
        # head is not supported
    ],
)
def test_static_activation_quantization(
    args, exp_min_val, exp_max_val, exp_quant, exp_loss
):
    """
    input = tensor([[ 0,  1,  2,  3,  4,  5]
                    [ 6,  7,  8,  9, 10, 11]])
    """
    # set up activation (and identity weight)
    batch_size, seq_len, input_size = 1, 2, 6
    input = torch.arange(
        (batch_size * seq_len * input_size), dtype=torch.bfloat16
    ).reshape((batch_size, seq_len, input_size))
    linear = torch.nn.Linear(input_size, input_size, bias=False)
    linear.weight.data = torch.eye(input_size, dtype=torch.bfloat16)

    # initialize quantization parameters
    scheme = QuantizationScheme(targets=[], input_activations=args)
    initialize_module_for_quantization(linear, scheme)
    assert getattr(linear, "quantization_scheme") is scheme
    linear.input_observer = MockMinMaxObserver("input", args, linear)

    # calibrate quantization parameters
    def calibrate_input_hook(_, args):
        if hasattr(linear, "input_global_scale"):
            global_scale = linear.input_observer.get_global_scale(args[0])
            linear.input_global_scale.data = global_scale

        if linear.quantization_scheme.input_activations.dynamic is False:
            scale, zero_point = linear.input_observer(args[0])
            linear.input_scale.data = scale
            linear.input_zero_point.data = zero_point

    linear.register_forward_pre_hook(calibrate_input_hook)

    # calibration forward pass
    output = linear(input)

    # check calibration
    if exp_min_val is not None:
        assert torch.equal(linear.input_observer.min_vals, exp_min_val)
    if exp_max_val is not None:
        assert torch.equal(linear.input_observer.max_vals, exp_max_val)

    # check forward pass
    assert torch.allclose(output, exp_quant.to(output.dtype))
    assert torch.nn.functional.mse_loss(output, input) <= exp_loss


class MockAttention(torch.nn.Module):
    pass


@pytest.mark.filterwarnings("ignore::UserWarning")
@pytest.mark.parametrize(
    "args,exp_min_val,exp_max_val,exp_quant,exp_loss",
    [
        (
            QuantizationArgs(
                num_bits=4,
                type="int",
                symmetric=True,
                strategy="tensor",
            ),
            torch.tensor([0.0]),
            torch.tensor([23.0]),
            torch.tensor(
                [
                    [
                        [
                            [0.0000, 0.0000, 3.0625, 3.0625],
                            [3.0625, 6.1250, 6.1250, 6.1250],
                            [9.1875, 9.1875, 9.1875, 12.2500],
                        ],
                        [
                            [12.2500, 12.2500, 15.3125, 15.3125],
                            [15.3125, 18.3750, 18.3750, 18.3750],
                            [21.5000, 21.5000, 21.5000, 21.5000],
                        ],
                    ]
                ]
            ),
            0.81,
        ),
        # static token is not supported
        # channel is not supported
        # group is not supported
        # tensor group is not supported
        # block is not supported
        (
            QuantizationArgs(
                num_bits=4,
                type="int",
                symmetric=True,
                strategy="attn_head",
            ),
            torch.tensor([[[0.0]], [[12.0]]]),
            torch.tensor([[[11.0]], [[23.0]]]),
            torch.tensor(
                [
                    [
                        [
                            [0.0000, 1.4688, 1.4688, 2.9375],
                            [4.4062, 4.4062, 5.8750, 7.3438],
                            [7.3438, 8.8125, 10.2500, 10.2500],
                        ],
                        [
                            [12.2500, 12.2500, 15.3125, 15.3125],
                            [15.3125, 18.3750, 18.3750, 18.3750],
                            [21.5000, 21.5000, 21.5000, 21.5000],
                        ],
                    ]
                ]
            ),
            0.55,
        ),
    ],
)
def test_static_attention_quantization(
    args, exp_min_val, exp_max_val, exp_quant, exp_loss
):
    """
    input = tensor([[[[ 0.,  1.,  2.,  3.],
                      [ 4.,  5.,  6.,  7.],
                      [ 8.,  9., 10., 11.]],

                     [[12., 13., 14., 15.],
                      [16., 17., 18., 19.],
                      [20., 21., 22., 23.]]]])
    """
    # set up attention
    batch_size, num_heads, seq_len, head_dim = 1, 2, 3, 4
    input = torch.arange(
        (batch_size * num_heads * seq_len * head_dim), dtype=torch.bfloat16
    ).reshape((batch_size, num_heads, seq_len, head_dim))
    attention = MockAttention()

    # initialize quantization parameters
    scheme = QuantizationScheme(targets=[], input_activations=args)
    initialize_qparams(
        attention, "k", args, (num_heads, None, head_dim), observed_dtype=torch.bfloat16
    )
    attention.quantization_scheme = scheme
    attention.quantization_status = QuantizationStatus.INITIALIZED
    attention.k_observer = MockMinMaxObserver("k", args, attention)

    # calibrate quantization parameters
    if scheme.input_activations.dynamic is False:
        scale, zero_point = attention.k_observer(input)
        attention.k_scale.data = scale
        attention.k_zero_point.data = zero_point

    # calibration forward pass
    output = forward_quantize(attention, input, "k", scheme.input_activations)

    # check calibration
    if exp_min_val is not None:
        assert torch.equal(attention.k_observer.min_vals, exp_min_val)
    if exp_max_val is not None:
        assert torch.equal(attention.k_observer.max_vals, exp_max_val)

    # check forward pass
    assert torch.allclose(output, exp_quant.to(output.dtype))
    assert torch.nn.functional.mse_loss(output, input) <= exp_loss



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import torch
from compressed_tensors import TRANSFORM_CONFIG_NAME
from compressed_tensors.transform import TransformConfig, TransformFactory
from compressed_tensors.utils.offload import has_offloaded_params


__all__ = ["apply_transform_config"]


def apply_transform_config(model: torch.nn.Module, config: TransformConfig):
    """
    Apply a transform config to a model. Weight transforms are fused into weights, while
    activation transforms are attached as submodules and trigger via pytorch hooks

    :param model: model to apply config to
    :param config: transform config to apply
    """
    for name, scheme in config.config_groups.items():
        factory = TransformFactory.from_scheme(scheme, name=name)
        factory.apply_to_model(model)

    # attach config to model for compression/serialization
    setattr(model, TRANSFORM_CONFIG_NAME, config)

    # ensure that tied weight transforms can be serialized without aliases
    # In the future, this could be done by transformers or model compressor
    # which would make this more robust to changing dispatches after transforms
    _tie_offloaded_tensors(model)


def _tie_offloaded_tensors(model: torch.nn.Module):
    """
    When accelerate replaces tensors with meta tensors during offloading, the meta
    tensors may not be identical, even if the offloaded values are identical.

    However, transformers can only serialize correctly if meta tensors are identical
    (see transformers#39263).

    This function collects all meta tensors which have shared offloaded values and sets
    those tensors to be identical so that they can be removed during serialization

    :param model: model potentially containing offloaded meta tensors to fix
    """

    # ensure that if a location shares an offloaded tensor pointers, that the
    # meta tensor is also identical (assigned to the first instance of parameter)
    ptr_to_meta: Dict[int, torch.nn.Parameter] = dict()
    for module in model.modules():
        if has_offloaded_params(module):
            for key, _ in module.named_parameters(recurse=False):
                offloaded_ptr = module._hf_hook.weights_map[key].data_ptr()

                if offloaded_ptr not in ptr_to_meta:
                    ptr_to_meta[offloaded_ptr] = getattr(module, key)
                setattr(module, key, ptr_to_meta[offloaded_ptr])



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationConfig,
    QuantizationScheme,
    QuantizationStatus,
    apply_quantization_config,
)
from compressed_tensors.quantization.lifecycle.forward import fake_quantize, quantize
from torch.nn import Linear


def create_config(bit_depth, quant_type, input_symmetry, weight_symmetry):
    weights = QuantizationArgs(
        num_bits=bit_depth, type=quant_type, symmetric=weight_symmetry
    )
    if input_symmetry is not None:
        inputs = QuantizationArgs(
            num_bits=bit_depth, type=quant_type, symmetric=input_symmetry
        )
    else:
        inputs = None

    config_groups = {
        "group_1": QuantizationScheme(
            targets=["Linear"], weights=weights, input_activations=inputs
        )
    }
    config = QuantizationConfig(
        config_groups=config_groups, quantization_status=QuantizationStatus.CALIBRATION
    )
    return config


@torch.no_grad
@pytest.mark.parametrize("bit_depth", [4, 8])
@pytest.mark.parametrize("quant_type", ["int"])
@pytest.mark.parametrize("input_symmetry", [True, False, None])
@pytest.mark.parametrize("weight_symmetry", [True, False])
def test_bit_depths(
    mock_per_tensor_calibration, bit_depth, quant_type, input_symmetry, weight_symmetry
):
    model = Linear(64, 64)
    quant_config = create_config(bit_depth, quant_type, input_symmetry, weight_symmetry)
    apply_quantization_config(model, quant_config)

    min = -1 * int(2**bit_depth / 2)
    max = int(2**bit_depth / 2) - 1

    inputs = torch.randn(32, 64)
    model.apply(
        lambda module: mock_per_tensor_calibration(
            module, base_name="weight", value=model.weight
        )
    )
    if input_symmetry is not None:
        model.apply(
            lambda module: mock_per_tensor_calibration(
                module, base_name="input", value=inputs
            )
        )
        assert model.input_zero_point >= min
        assert model.input_zero_point <= max

        input_max = torch.max(inputs)
        input_min = torch.min(inputs)
        diff_from_max = abs(
            abs(model.input_scale * (max - model.input_zero_point)) - abs(input_max)
        )
        diff_from_min = abs(
            abs(model.input_scale * abs(min - model.input_zero_point)) - abs(input_min)
        )
        assert diff_from_max < model.input_scale or diff_from_min < model.input_scale

    assert model.weight_zero_point >= min
    assert model.weight_zero_point <= max

    weight_max = torch.max(model.weight)
    weight_min = torch.min(model.weight)
    diff_from_max = abs(
        abs(model.weight_scale * (max - model.weight_zero_point)) - abs(weight_max)
    )
    diff_from_min = abs(
        abs(model.weight_scale * abs(min - model.weight_zero_point)) - abs(weight_min)
    )
    assert diff_from_max < model.weight_scale or diff_from_min < model.weight_scale

    quantized_weight = fake_quantize(
        model.weight,
        model.weight_scale,
        model.weight_zero_point,
        model.quantization_scheme.weights,
    )
    assert not torch.any(quantized_weight < min).item()
    assert not torch.any(quantized_weight > max).item()


@torch.no_grad
@pytest.mark.parametrize("bit_depth", [8])
@pytest.mark.parametrize("quant_type", ["float"])
@pytest.mark.parametrize("input_symmetry", [True, False, None])
@pytest.mark.parametrize("weight_symmetry", [True, False])
def test_fp8(
    mock_per_tensor_calibration, bit_depth, quant_type, input_symmetry, weight_symmetry
):
    model = Linear(64, 64)
    quant_config = create_config(bit_depth, quant_type, input_symmetry, weight_symmetry)
    apply_quantization_config(model, quant_config)

    dtype_info = torch.finfo(torch.float8_e4m3fn)
    min = dtype_info.min
    max = dtype_info.max

    inputs = torch.randn(32, 64)
    model.apply(
        lambda module: mock_per_tensor_calibration(
            module, base_name="weight", value=model.weight
        )
    )
    assert model.weight_zero_point.dtype == torch.float8_e4m3fn
    model.weight_zero_point.data = model.weight_zero_point.to(model.weight.dtype)
    if input_symmetry is not None:
        model.apply(
            lambda module: mock_per_tensor_calibration(
                module, base_name="input", value=inputs
            )
        )
        assert model.input_zero_point.dtype == torch.float8_e4m3fn
        model.input_zero_point.data = model.input_zero_point.to(model.weight.dtype)
        assert model.input_zero_point >= min
        assert model.input_zero_point <= max

        inputs_fake_quant = quantize(
            inputs,
            model.input_scale,
            model.input_zero_point,
            model.quantization_scheme.input_activations,
        )
        input_max = torch.max(inputs_fake_quant)
        input_min = torch.min(inputs_fake_quant)
        diff_from_max = abs(input_max - max)
        diff_from_min = abs(input_min - min)
        assert diff_from_max.item() == 0.0 or diff_from_min.item() == 0.0

    assert model.weight_zero_point >= min
    assert model.weight_zero_point <= max

    weight_fake_quant = quantize(
        model.weight,
        model.weight_scale,
        model.weight_zero_point,
        model.quantization_scheme.weights,
    )
    weight_max = torch.max(weight_fake_quant)
    weight_min = torch.min(weight_fake_quant)
    diff_from_max = abs(weight_max - max)
    diff_from_min = abs(weight_min - min)
    assert diff_from_max.item() == 0.0 or diff_from_min.item() == 0.0



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End-to-end tests for asymmetric quantization with zero-point decompression.
"""

import pytest
import torch
from compressed_tensors.compressors.model_compressors.model_compressor import (
    ModelCompressor,
)
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationConfig,
    QuantizationScheme,
    QuantizationStrategy,
    apply_quantization_config,
)
from torch.nn import Linear, Module


class SimpleModel(Module):
    """Simple model for testing"""

    def __init__(self, input_dim=512, hidden_dim=256, output_dim=128):
        super().__init__()
        self.layer1 = Linear(input_dim, hidden_dim, bias=False)
        self.layer2 = Linear(hidden_dim, output_dim, bias=False)

    def forward(self, x):
        x = self.layer1(x)
        x = torch.relu(x)
        x = self.layer2(x)
        return x


def create_asymmetric_quant_config(
    num_bits=4, strategy=QuantizationStrategy.GROUP, group_size=128
) -> QuantizationConfig:
    """Create an asymmetric quantization config"""
    config_groups = {
        "group_1": QuantizationScheme(
            targets=["Linear"],
            weights=QuantizationArgs(
                num_bits=num_bits,
                strategy=strategy.value,
                group_size=(
                    group_size if strategy == QuantizationStrategy.GROUP else None
                ),
                symmetric=False,
            ),
        ),
    }
    return QuantizationConfig(config_groups=config_groups)


@pytest.mark.parametrize(
    "strategy,group_size",
    [
        (QuantizationStrategy.GROUP, 128),
        (QuantizationStrategy.CHANNEL, None),
    ],
)
def test_end_to_end_asymmetric_quantization(
    strategy,
    group_size,
    mock_per_group_calibration,
    mock_per_channel_calibration,
):
    """
    Test end-to-end workflow: quantize -> compress -> decompress in memory
    """
    model = SimpleModel()
    original_weights = {
        "layer1": model.layer1.weight.detach().clone(),
        "layer2": model.layer2.weight.detach().clone(),
    }

    quant_config = create_asymmetric_quant_config(
        num_bits=4, strategy=strategy, group_size=group_size
    )
    # Set pack-quantized format for ModelCompressor usage
    quant_config.format = CompressionFormat.pack_quantized.value
    apply_quantization_config(model, quant_config)

    if strategy == QuantizationStrategy.GROUP:
        mock_per_group_calibration(
            model.layer1, "weight", model.layer1.weight, group_size
        )
        mock_per_group_calibration(
            model.layer2, "weight", model.layer2.weight, group_size
        )
    else:
        mock_per_channel_calibration(model.layer1, "weight", model.layer1.weight)
        mock_per_channel_calibration(model.layer2, "weight", model.layer2.weight)

    # Compress and decompress in memory using ModelCompressor
    mc = ModelCompressor(quantization_config=quant_config)
    mc.compress_model(model)

    # Verify compression created zero-point parameters
    assert hasattr(model.layer1, "weight_zero_point")
    assert hasattr(model.layer2, "weight_zero_point")
    assert model.layer1.weight_zero_point.dtype == torch.int32
    assert model.layer2.weight_zero_point.dtype == torch.int32

    # Decompress in memory
    mc.decompress_model(model)

    # Verify decompression restored weights correctly
    assert model.layer1.weight.shape == original_weights["layer1"].shape
    assert model.layer2.weight.shape == original_weights["layer2"].shape
    assert model.layer1.weight.dtype.is_floating_point
    assert model.layer2.weight.dtype.is_floating_point
    assert not torch.isnan(model.layer1.weight).any()
    assert not torch.isnan(model.layer2.weight).any()
    assert not torch.isinf(model.layer1.weight).any()
    assert not torch.isinf(model.layer2.weight).any()


@pytest.mark.parametrize("num_bits", [4, 8])
def test_asymmetric_quantization_accuracy(num_bits, mock_per_group_calibration):
    """
    Test that asymmetric quantization with zero-point preserves accuracy better
    than symmetric quantization for biased weight distributions.
    """
    shape = (256, 512)
    biased_weights = torch.randn(shape) + 2.0

    quant_config = create_asymmetric_quant_config(
        num_bits=num_bits,
        strategy=QuantizationStrategy.GROUP,
        group_size=128,
    )
    quant_config.format = CompressionFormat.pack_quantized.value

    class SingleLayer(Module):
        def __init__(self):
            super().__init__()
            self.layer = Linear(shape[1], shape[0], bias=False)

    model = SingleLayer()
    apply_quantization_config(model, quant_config)

    with torch.no_grad():
        model.layer.weight.copy_(biased_weights)
    mock_per_group_calibration(model.layer, "weight", model.layer.weight, 128)

    # Compress and decompress in memory using ModelCompressor
    mc = ModelCompressor(quantization_config=quant_config)
    mc.compress_model(model)
    mc.decompress_model(model)

    decompressed_weights = model.layer.weight
    assert decompressed_weights.shape == shape
    assert not torch.isnan(decompressed_weights).any()
    assert not torch.isinf(decompressed_weights).any()
    threshold = torch.std(torch.rand(shape) - torch.rand(shape))
    assert torch.std(biased_weights - decompressed_weights) < threshold



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re
from collections import defaultdict
from collections.abc import Generator
from typing import Iterable, List, Mapping, Optional, Tuple, Union

import torch
from compressed_tensors.utils.internal import InternalModule


_LOGGER: logging.Logger = logging.getLogger(__name__)


__all__ = [
    "match_named_modules",
    "match_named_parameters",
    "match_targets",
    "match_modules_set",
    "get_lowest_common_ancestor_name",
    "is_match",
    "is_narrow_match",
]


FusedMappping = Mapping[str, Iterable[str]]


def match_named_modules(
    model: torch.nn.Module,
    targets: Optional[Iterable[str]],
    ignore: Optional[Iterable[str]] = None,
    fused: Optional[FusedMappping] = None,
    warn_on_fail: bool = False,
) -> Generator[Tuple[str, torch.nn.Module]]:
    """
    Yields names and modules which match `targets` but do not match `ignore`.
    Values are returned in order of `model.named_modules()`

    :param model: model containing submodules to match against
    :param targets: target strings, potentially containing "re:" prefixes
    :param ignore: targets to ignore, potentially containing "re:" prefixes
    :fused: optional mapping from suffixes of fused modules to the suffixes of their
        corresponding shards. See `compressed_tensors.utils.match.is_match`
    :param warn_on_fail: if True, warns if any targets do not match any modules in model
    :return: generator of module names and modules
    """
    targets = targets or []
    ignore = ignore or []

    unmatched_targets = set(targets)

    for name, module in model.named_modules():
        for target in targets:
            if is_match(name, module, target, fused=fused):
                unmatched_targets -= {target}
                if not is_match(name, module, ignore, fused=fused):
                    yield name, module
                break

    if warn_on_fail:
        for target in unmatched_targets:
            _LOGGER.warning(
                f"Could not match `{target}` in instance of {model.__class__.__name__}"
            )


def match_named_parameters(
    model: torch.nn.Module,
    targets: Optional[Iterable[str]],
    ignore: Optional[Iterable[str]] = None,
    fused: Optional[FusedMappping] = None,
    warn_on_fail: bool = False,
) -> Generator[Tuple[str, torch.nn.Module, torch.nn.Parameter]]:
    """
    Yields parameters which match `targets` but do not match `ignore`.
    Values are returned in order of `model.named_modules()`

    :param model: model containing params to match against
    :param targets: target strings, potentially containing "re:" prefixes
    :param ignore: targets to ignore, potentially containing "re:" prefixes
    :fused: optional mapping from suffixes of fused modules to the suffixes of their
        corresponding shards. See `compressed_tensors.utils.match.is_match`
    :param warn_on_fail: if True, warns if any targets do not match any params in model
    :return: generator of fully-qualified param names, parent modules, and params
    """
    targets = targets or []
    ignore = ignore or []

    unmatched_targets = set(targets)
    for module_name, module in model.named_modules():
        if isinstance(module, InternalModule):
            continue

        for param_name, param in module.named_parameters(recurse=False):
            param_fqn = f"{module_name}.{param_name}"
            for target in targets:
                if _match_name(param_fqn, target, fused):
                    unmatched_targets -= {target}

                    if not any(_match_name(param_fqn, ign, fused) for ign in ignore):
                        yield param_fqn, module, param

    if warn_on_fail:
        for target in unmatched_targets:
            _LOGGER.warning(
                f"Could not match `{target}` in instance of {model.__class__.__name__}"
            )


def match_targets(
    name: str, module: torch.nn.Module, targets: Optional[Iterable[str]]
) -> List[str]:
    """
    Returns the targets that match the given name and module.

    :param name: the name of the module
    :param module: the module to match
    :param targets: the target strings, potentially containing "re:" prefixes
    :return: the targets that match the given name and module

    Outputs are ordered by type: exact name match, regex name match, class name match
    """
    targets = targets or []

    if isinstance(module, InternalModule):
        return []

    # The order of the output `matches` list matters, they are arranged from most
    # specific to least specific, and this order will be used when merging configs.
    # The entries are sorted in the following order:
    #     1. matches on exact strings
    #     2. matches on regex patterns
    #     3. matches on module names (e.g. "Linear")

    targets = sorted(targets, key=lambda x: ("re:" in x, x))
    matched_targets = []
    for target in targets:
        if _match_name(name, target):
            matched_targets.append(target)

    for target in targets:
        if _match_class(module, target) and target not in matched_targets:
            matched_targets.append(target)

    return matched_targets


def get_lowest_common_ancestor_name(names: list[str | None]) -> str:
    """
    Given a list of names, returns the lowest-scope common name ignoring Nones.

    Implementation is a small alteration of os.path.commonprefix
    https://docs.python.org/3/library/os.path.html#os.path.commonprefix

    ([s1, s2]->prefix->result)
    # case 0: multiple modules: [abc.a., abc.b.] -> .abc. -> abc
    # case 1: single module: [abc.] -> .abc. -> abc
    # case 2: substring modules: [abc., ab.] -> .ab -> ""
    # case 3: parent & child: [ab., ab.a.] -> .ab. -> ab
    """
    names = [name for name in names if name is not None]
    if len(names) == 0:
        return ""

    # 1) find longest shared prefix
    s1 = "." + min(names) + "."
    s2 = "." + max(names) + "."
    common_prefix = os.path.commonprefix([s1, s2])
    # 2) throw away right most dot and name fragment, throw away leftmost char
    # ".keep.thro" -> "keep", "." -> ""
    return common_prefix[1 : common_prefix.rfind(".")]


def match_modules_set(
    model: torch.nn.Module,
    targets: Optional[Iterable[str]],
    ignore: Optional[Iterable[str]] = None,
    error_on_module_rematch: bool = True,
) -> Generator[List[List[torch.nn.Module]]]:
    """
    Yields modules grouped by parent context.

    We group by parent context so that we can return ALL matches of a
    specific target that can be paired with another target. This is most
    relevant in the case of MoE modules with multiple modules for each
    expert i.e. post_attention_layernorm <-> mlp.expert.N.gate_proj,
    mlp.expert.N.up_proj for all N. The parent context will differ from
    one layer to another while being the same for one expert to another.

    Each returned group is a list (of lists) with the same size
    and order as `targets` while all matches for each target and
    the overall order of the groups are ordered in the same way
    as `model.named_modules`


    E.g. the following targets would yield modules belonging to the following layers:
    ```python3
    match_modules_set(model, ["q_proj", "k_proj", "v_proj"]) == (
        [
            [`layers.0.self_attn.q_proj`],
            [`layers.0.self_attn.k_proj`],
            [`layers.0.self_attn.v_proj`],
        ],
        [
            [`layers.1.self_attn.q_proj`],
            [`layers.1.self_attn.k_proj`],
            [`layers.1.self_attn.v_proj`],
        ],
        ...
    )
    ```

    This can be used to match layers to their corresponding downstream counterparts.
    For example, matching layer norms to their subsequent linear layers
    ```python3
    for norm, q, k, v in match_modules_set(model, (norm_tgt, q_tgt, k_tgt, v_tgt)):
        fuse_norm_linears(*norm, [*q, *k, *v])
    ```

    Alternatively for MoE you would get multiple matches
    per target per group, E.g.

    ```python3

    targets = [
        "post_attention_layernorm",
        "up_proj",
        "down_proj"
    ]
    match_modules_set(model, targets) == (
        [
            [layers.0.post_attention_layernorm],
            [
                `layers.0.mlp.experts.0.up_proj`,
                `layers.0.mlp.experts.1.up_proj`,
                ...
            ],
            [
                `layers.0.mlp.experts.0.down_proj`,
                `layers.0.mlp.experts.1.down_proj`,
                ...

            ]
        ], # <- first yield
        [
            [layers.1.post_attention_layernorm],
            [
                `layers.1.mlp.experts.0.up_proj`,
                `layers.1.mlp.experts.1.up_proj`,
                ...
            ],
            [
                `layers.1.mlp.experts.0.down_proj`,
                `layers.1.mlp.experts.1.down_proj`,
                ...
            ]
        ],
        ...
    )
    ```

    :param model: model containing modules to match against
    :param targets: target strings, potentially containing "re:" prefixes
    :param ignore: targets to ignore, potentially containing "re:" prefixes
    :param error_on_module_rematch: if True, errors when a module gets
      matched to multiple targets, if False, no error. (Defaults to True)
    """
    targets = targets or []
    ignore = ignore or []

    # as we iterate through modules and try to match them with targets,
    # the algorithm can be in 2 possible states:
    # 0) unmatched_targets > 0, i.e. some of the targets haven't been matched.
    #   Keep matching until all targets have at least one match
    # 1) unmatched_targets == 0 i.e. we have at least one match for each target.
    #   At this point we are unsure if we have a full set or if we need to add
    #   more matches.
    # There are 3 things that can happen once were in state 1:
    # A) found a new match with same parent_context,
    #   (add it to matches and keep going)
    # B) found a new match with different parent_context, i.e. we found a match
    #   that requires a deeper parent context, this indicates that this match
    #   should be part of a new set.
    #   (yield current set [not including newest match] and go back to state 0)
    # C) ran out of modules, we will always yield the final remaining set when
    #   we we've iterated through all the modules in the model.
    #   (yield final set then exit.)
    # Note: its possible to iterate through all the modules in the model while
    #   not having a full matched set if the user specified a bad matching, in
    #   that case something has gone wrong and we error
    matches = defaultdict(list)
    parent_context = None
    unmatched_targets = set(targets)

    for name, module in model.named_modules():
        matched_targets_for_cur_module = set()
        for target in targets:
            if is_match(name, module, target, ignore):
                new_parent_context = get_lowest_common_ancestor_name(
                    [name, parent_context]
                )

                # code for (B)
                if not unmatched_targets and new_parent_context != parent_context:
                    yield [matches[target] for target in targets]
                    matches = defaultdict(list)
                    new_parent_context = name
                    unmatched_targets = set(targets)

                matches[target].append(module)
                parent_context = new_parent_context
                unmatched_targets -= {target}
                matched_targets_for_cur_module |= {target}

        if len(matched_targets_for_cur_module) > 1 and error_on_module_rematch:
            raise ValueError(
                f"module: {name} was matched with multiple targets: "
                f"{matched_targets_for_cur_module} which is unexpected "
                "disable this check by setting `error_on_module_rematch = False`"
            )

    # never found anything
    if unmatched_targets == set(targets):
        return

    # code for (C)
    if not unmatched_targets:  # have a full matching
        yield [matches[target] for target in targets]
        return

    raise ValueError(
        f"Found a final incomplete set with matches found for keys: "
        f"{set(targets) - unmatched_targets} "
        f"but no matches found for keys: {unmatched_targets}"
    )


def is_match(
    name: str,
    module: torch.nn.Module,
    targets: Union[str, Iterable[str]],
    ignore: Union[str, Iterable[str]] = tuple(),
    fused: Optional[FusedMappping] = None,
) -> bool:
    """
    Returns true if either module name or module parent classes match against target
    and the module is not an internal module. The name and module may refer to a fused
    module defined by vLLM. In these cases, a `fused` mapping must be provided.

    For example, in `vllm/model_executor/models/llama.py`:
    ```python
    packed_modules_mapping = {
        "qkv_proj": ["q_proj", "k_proj", "v_proj"],
        "gate_up_proj": ["gate_proj", "up_proj"]
    }
    ```

    :param name: name of module
    :param module: module to match
    :param target: target which matches name or module, potentially contains regex
    :fused: optional mapping from suffixes of fused modules to the suffixes of their
        corresponding shards
    """
    targets = [targets] if isinstance(targets, str) else targets
    ignore = [ignore] if isinstance(ignore, str) else ignore

    return not isinstance(module, InternalModule) and (
        any(
            _match_name(name, target, fused) or _match_class(module, target)
            for target in targets
        )
        and not any(
            _match_name(name, ign, fused) or _match_class(module, ign) for ign in ignore
        )
    )


def is_narrow_match(
    model: torch.nn.Module,
    targets: Union[str, Iterable[str]],
    name: str,
    module: Optional[torch.nn.Module] = None,
) -> bool:
    """
    Checks if any of the targets narrowly match the module. A target narrowly matches
    a module if the target matches the module, but does not match the module's parent

    :param model: model containing both module and its parent
    :param targets: target strings, potentially containing "re:" prefixes
    :param name: name of module to match
    :param module: module to match. If none is provided, then get module from model
    :return: True if any of the targets narrow match the module
    """
    targets = [targets] if isinstance(targets, str) else targets
    module = module if module is not None else model.get_submodule(name)

    parent_name = name.rsplit(".", 1)[0]
    parent = model.get_submodule(parent_name)

    return any(
        is_match(name, module, target) and not is_match(parent_name, parent, target)
        for target in targets
    )


def _match_name(name: str, target: str, fused: Optional[FusedMappping] = None) -> bool:
    """
    Returns true if target string begins with "re:" and regex matches or if target
    string exactly matches name. If the name refers to a fused module defined by vLLM,
    a `fused` mapping must be provided.

    :param name: name of module
    :param target: target name, potentially contains regex
    :fused: optional mapping from suffixes of fused modules to the suffixes of their
        corresponding shards
    """
    if fused is not None:
        for fused_suffix in fused:
            if name.endswith(fused_suffix):
                name_stripped = name.removesuffix(fused_suffix)
                return any(
                    _match_name(name_stripped + shard_suffix, target)
                    for shard_suffix in fused[fused_suffix]
                )

    if target.startswith("re:"):
        return re.match(target.removeprefix("re:"), name) is not None
    else:
        return target == name


def _match_class(module: torch.nn.Module, target: str) -> bool:
    """
    Returns true if any torch parent class names match the target string exactly.
    A special exception is made for vllm's `LinearBase` class which matches `Linear`

    :param module: module to match
    :param target: target which matches name or module
    """
    # will never match against a regex pattern since `:` is not allowed in class names
    return any(
        (
            issubclass(cls, torch.nn.Module)
            and (
                cls.__name__ == target
                or (cls.__name__ == "LinearBase" and target == "Linear")
            )
        )
        for cls in module.__class__.__mro__
    )



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import warnings
from functools import wraps
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    TypeVar,
)

import numpy
import torch
from transformers import AutoConfig, PretrainedConfig


T = TypeVar("T", bound="Callable")  # used by `deprecated`


if TYPE_CHECKING:
    from compressed_tensors.compressors import ModelCompressor


__all__ = [
    "infer_compressor_from_model_config",
    "fix_fsdp_module_name",
    "tensor_follows_mask_structure",
    "replace_module",
    "is_compressed_tensors_config",
    "getattr_chain",
    "deprecated",
    "Aliasable",
    "combine_shards",
    "shard_tensor",
    "pack_bitmasks",
    "unpack_bitmasks",
    "patch_attr",
    "patch_attrs",
    "ParameterizedDefaultDict",
    "get_num_attn_heads",
    "get_num_kv_heads",
    "get_head_dim",
]

FSDP_WRAPPER_NAME = "_fsdp_wrapped_module"


def infer_compressor_from_model_config(
    pretrained_model_name_or_path: str,
) -> Optional["ModelCompressor"]:  # noqa: F821
    """
    Given a path to a model config, extract a sparsity config if it exists and return
    the associated ModelCompressor

    :param pretrained_model_name_or_path: path to model config on disk or HF hub
    :return: matching compressor if config contains a sparsity config
    """
    from compressed_tensors.compressors import ModelCompressor
    from compressed_tensors.config import CompressionConfig

    config = AutoConfig.from_pretrained(pretrained_model_name_or_path)
    sparsity_config = ModelCompressor.parse_sparsity_config(config)
    if sparsity_config is None:
        return None

    format = sparsity_config.get("format")
    sparsity_config = CompressionConfig.load_from_registry(format, **sparsity_config)
    compressor = ModelCompressor.load_from_registry(format, config=sparsity_config)
    return compressor


def fix_fsdp_module_name(name: str) -> str:
    """
    Remove FSDP wrapper prefixes from a module name
    Accounts for scenario where FSDP_WRAPPER_NAME is
    at the end of the name, as well as in the middle.
    :param name: name to strip
    :return: stripped name
    """
    return name.replace(FSDP_WRAPPER_NAME + ".", "").replace(
        "." + FSDP_WRAPPER_NAME, ""
    )


def tensor_follows_mask_structure(tensor, mask: str = "2:4") -> bool:
    """
    :param tensor: tensor to check
    :param mask: mask structure to check for, in the format "n:m"
    :return: True if the tensor follows the mask structure, False otherwise.
        Note, some weights can incidentally be zero, so we check for
        atleast n zeros in each chunk of size m
    """

    n, m = tuple(map(int, mask.split(":")))
    # Reshape the tensor into chunks of size m
    tensor = tensor.view(-1, m)

    # Count the number of zeros in each chunk
    zero_counts = (tensor == 0).sum(dim=1)

    # Check if the number of zeros in each chunk atleast n
    # Greater than sign is needed as some weights can incidentally
    # be zero
    if not torch.all(zero_counts >= n).item():
        raise ValueError()

    return True


def replace_module(model: torch.nn.Module, name: str, new_module: torch.nn.Module):
    if "." in name:
        parent_name = name.rsplit(".", 1)[0]
        child_name = name[len(parent_name) + 1 :]
        parent = model.get_submodule(parent_name)
    else:
        parent_name = ""
        parent = model
        child_name = name
    setattr(parent, child_name, new_module)


def is_compressed_tensors_config(compression_config: Any) -> bool:
    """
    Returns True if CompressedTensorsConfig is available from transformers and
    compression_config is an instance of CompressedTensorsConfig

    See: https://github.com/huggingface/transformers/pull/31704
    """
    try:
        from transformers.utils.quantization_config import CompressedTensorsConfig

        return isinstance(compression_config, CompressedTensorsConfig)
    except ImportError:
        return False


def getattr_chain(obj: Any, chain_str: str, *args, **kwargs) -> Any:
    """
    Chain multiple getattr calls, separated by `.`

    :param obj: base object whose attributes are being retrieved
    :param chain_str: attribute names separated by `.`
    :param default: default value, throw error otherwise
    """
    if len(args) >= 1:
        has_default = True
        default = args[0]
    elif "default" in kwargs:
        has_default = True
        default = kwargs["default"]
    else:
        has_default = False

    attr_names = chain_str.split(".")

    res = obj
    for attr_name in attr_names:
        if not hasattr(res, attr_name):
            if has_default:
                return default
            else:
                raise AttributeError(f"{res} object has no attribute {attr_name}")
        res = getattr(res, attr_name)

    return res


def deprecated(
    future_name: Optional[str] = None, message: Optional[str] = None
) -> Callable[[T], T]:
    """
    Decorator to mark functions as deprecated

    :param new_function: Function called in place of deprecated function
    :param message: Deprecation message, replaces default deprecation message
    """

    def decorator(func: T) -> T:
        nonlocal message

        if message is None:
            message = (
                f"{func.__name__} is deprecated and will be removed in a future release"
            )
            if future_name is not None:
                message += f". Please use {future_name} instead."

        @wraps(func)
        def wrapped(*args, **kwargs):
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapped

    return decorator


class Aliasable:
    """
    A mixin for enums to allow aliasing of enum members

    Example:
    >>> class MyClass(Aliasable, int, Enum):
    >>>     ...
    """

    @staticmethod
    def get_aliases() -> Dict[str, str]:
        raise NotImplementedError()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            aliases = self.get_aliases()
            return self.value == other.value or (
                aliases.get(self.value, self.value)
                == aliases.get(other.value, other.value)
            )
        else:
            aliases = self.get_aliases()
            self_value = aliases.get(self.value, self.value)
            other_value = aliases.get(other, other)
            return self_value == other_value

    def __hash__(self):
        canonical_value = self.aliases.get(self.value, self.value)
        return hash(canonical_value)


def shard_tensor(
    tensor: torch.Tensor, shard_sizes: List[int], dim: int = 0
) -> List[torch.Tensor]:
    """
    Shards a tensor into a list of tensors along a given dimension.

    raises: ValueError: If the sum of shard_sizes does not match the
        size of the tensor along the given dimension.

    :param tensor: The input tensor to shard.
    :param shard_sizes : List of sizes for each shard along the specified dimension.
    :param dim : The dimension along which to shard the tensor.
    :returns: A list of tensors sharded along the specified dimension.
    """
    if sum(shard_sizes) != tensor.size(dim):
        raise ValueError(
            "Sum of shard_sizes must equal the size of the tensor "
            "along the specified dimension."
        )

    shards = []
    start_idx = 0

    for size in shard_sizes:
        end_idx = start_idx + size
        shard = tensor.narrow(dim, start_idx, size)
        shards.append(shard)
        start_idx = end_idx

    return shards


def combine_shards(shards, dim=0):
    """
    Combine decompressed shards along a given dimension using `narrow`.

    :param shards: List of decompressed shard tensors.
    :param dim: Dimension to combine along (default: 0).
    :return: Combined decompressed tensor.
    """
    if not shards:
        raise ValueError("The list of shards is empty.")

    # Assert that all shards have the same dtype
    shard_dtypes = {shard.dtype for shard in shards}
    if len(shard_dtypes) > 1:
        raise ValueError("All shards must have the same dtype.")

    # Determine the total shape of the combined tensor
    total_shape = list(shards[0].shape)
    total_shape[dim] = sum(shard.shape[dim] for shard in shards)

    # Create the combined tensor
    combined = torch.zeros(total_shape, dtype=shards[0].dtype, device=shards[0].device)

    # Fill the combined tensor using narrow
    shard_offset = 0
    for shard in shards:
        shard_size = shard.shape[dim]
        combined.narrow(dim, shard_offset, shard_size).copy_(shard)
        shard_offset += shard_size

    return combined


def pack_bitmasks(bytemasks: torch.Tensor) -> torch.Tensor:
    """
    Converts a bytemask tensor to a bitmask tensor to reduce memory. Shape RxC will be
    compressed to R x ceil(C/8)

    :param bytemasks: mask tensor where each byte corresponds to a weight
    :return: mask tensor where each bit corresounds to a weight
    """
    packed_bits_numpy = numpy.packbits(bytemasks.numpy(), axis=-1, bitorder="little")
    packed_bits_torch = torch.from_numpy(packed_bits_numpy)

    return packed_bits_torch


def unpack_bitmasks(
    packed_bitmasks: torch.Tensor, original_shape: List[int]
) -> torch.Tensor:
    """
    Converts a bitmask tensor back to a bytemask tensor for use during decompression

    :param packed_bitmasks: mask tensor where each bit corresponds to a weight
    :param original_shape: dense shape to decompress to
    :return: boolean mask of weights in the original dense shape
    """
    # Unpack the bits
    unpacked_bits = numpy.unpackbits(
        packed_bitmasks.cpu().numpy(),
        axis=-1,
        count=original_shape[-1],
        bitorder="little",
    )

    # Reshape to match the original shape
    unpacked_bitmasks_torch = torch.from_numpy(
        unpacked_bits.reshape(original_shape).astype(bool)
    )

    return unpacked_bitmasks_torch


@contextlib.contextmanager
def patch_attr(base: object, attr: str, value: Any):
    """
    Patch the value of an object attribute. Original value is restored upon exit

    :param base: object which has the attribute to patch
    :param attr: name of the the attribute to patch
    :param value: used to replace original value

    Usage:
    >>> from types import SimpleNamespace
    >>> obj = SimpleNamespace()
    >>> with patch_attr(obj, "attribute", "value"):
    ...     assert obj.attribute == "value"
    >>> assert not hasattr(obj, "attribute")
    """
    _sentinel = object()
    original_value = getattr(base, attr, _sentinel)

    setattr(base, attr, value)
    try:
        yield
    finally:
        if original_value is not _sentinel:
            setattr(base, attr, original_value)
        else:
            delattr(base, attr)


@contextlib.contextmanager
def patch_attrs(bases: Iterable[Any], attr: str, values: Iterable[Any]):
    """
    Same as `patch_attr` but for a list of objects to patch
    Patch attribute for a list of objects with list of values.
    Original values are restored upon exit

    :param bases: objects which has the attribute to patch
    :param attr: name of the the attribute to patch
    :param values: used to replace original values. Must be same
        length as bases

    Usage:
    >>> from types import SimpleNamespace
    >>> obj1 = SimpleNamespace()
    >>> obj2 = SimpleNamespace()
    >>> with patch_attr([obj1, obj2], "attribute", ["value1", "value2"]):
    ...     assert obj1.attribute == "value1"
    ...     assert obj2.attribute == "value2"
    >>> assert not hasattr(obj1, "attribute")
    >>> assert not hasattr(obj2, "attribute")
    """
    with contextlib.ExitStack() as stack:
        for base, value in zip(bases, values):
            stack.enter_context(patch_attr(base, attr, value))
        yield


class ParameterizedDefaultDict(dict):
    """
    Similar to `collections.DefaultDict`, but upon fetching a key which is missing,
    the key is passed as arguments to the `default_factory`

    :param default_factory: function which takes a key as input and returns the
        corresponding default value
    """

    def __init__(self, default_factory: Callable[[Any], Any]):
        self.default_factory = default_factory
        self._factory_kwargs = MappingProxyType({})

    def __missing__(self, key: Any) -> Any:
        if isinstance(key, tuple):
            value = self.default_factory(*key, **self._factory_kwargs)
        else:
            value = self.default_factory(key, **self._factory_kwargs)
        self[key] = value
        return value

    def get(self, *args, factory_kwargs: Mapping = MappingProxyType({})) -> Any:
        """
        Similar to `__getitem__`, but allows passing kwargs to factory function

        :param \\*args: args whose tuple will value will be treated as key
        :param factory_kwargs: keyword arguments to pass to `default_factory`
        :return: dictionary entry for given key
        """
        with patch_attr(self, "_factory_kwargs", factory_kwargs):
            return self[args]


def get_num_attn_heads(config: PretrainedConfig) -> int:
    """
    Get the number of attention heads used by a model

    :param config: model config
    :return: num_attention_heads of model
    """
    if hasattr(config, "num_attention_heads"):
        return config.num_attention_heads

    elif hasattr(config, "hidden_size") and hasattr(config, "head_dim"):
        return config.hidden_size // config.head_dim

    else:
        raise ValueError(
            "Cannot determine num_attention_heads from config. Config must define "
            "either `num_attention_heads` or both `hidden_size` and `head_dim`. "
            f"{config}"
        )


def get_num_kv_heads(config: PretrainedConfig) -> int:
    """
    Get the number of key-value attention heads used by a model

    :param config: model config
    :return: num_key_value_heads of model
    """
    if hasattr(config, "num_key_value_heads"):
        return config.num_key_value_heads

    else:
        raise ValueError(
            "Cannot determine num_key_value_heads from config. Config must define "
            f"`num_key_value_heads`. {config}"
        )


def get_head_dim(config: PretrainedConfig) -> int:
    """
    Get the number of dimensions used by the attention heads of a model

    :param config: model config
    :return: head_dim of model
    """
    if hasattr(config, "head_dim"):
        return config.head_dim

    elif hasattr(config, "hidden_size") and hasattr(config, "num_attention_heads"):
        return config.hidden_size // config.num_attention_heads

    else:
        raise ValueError(
            "Cannot determine head_dim from config. Config must define "
            "either `head_dim` or both `hidden_size` and `num_attention_heads`. "
            f"{config}"
        )



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Any, Callable, Dict, List, Optional, Tuple
from weakref import ReferenceType, ref

from compressed_tensors.quantization.lifecycle.forward import forward_quantize
from compressed_tensors.utils import getattr_chain
from compressed_tensors.utils.internal import InternalModule
from torch import Tensor
from torch.nn import Module
from torch.utils.hooks import RemovableHandle
from transformers import Cache, PretrainedConfig, PreTrainedModel


__all__ = [
    "QuantizedKVCache",
    "initialize_hooked_kv_cache",
    "register_key_hook",
    "register_value_hook",
    "KV_CACHE_ATTR",
]


KV_CACHE_ATTR = "kv_cache"


class QuantizedKVCache(InternalModule):
    """
    QuantizedKVCache module which wraps the functionality of any existing kvcache args.
    Unlike transform Cache instances, this cache is a `torch.nn.Module` which can be
    hooked to trigger transforms and calibration hooks.

    This module works by being registered as a submodule to attention modules via
    `initialize_hooked_kv_cache`, then adding a hook which replaces `past_key_values`
    kwargs with this module. This module adopts the functionality of the replaced cache,
    preserving caching functionality such as sliding window attention, ect.

    :param attn_module: parent attention module
    """

    def __init__(self, config: PretrainedConfig, attn_module: Module):
        super().__init__()
        self.config = config
        self.attn_module = ref(attn_module)  # avoid circular reference
        self.past_key_values: Optional[ReferenceType[Cache]] = None

    def update(self, *args, **kwargs) -> Tuple[Tensor, Tensor]:
        return self(*args, **kwargs)

    def forward(
        self,
        key_states: Tensor,
        value_states: Tensor,
        *args,
        **kwargs,
    ) -> Tuple[Tensor, Tensor]:
        # quantization
        module = self.attn_module()
        quant_args_attr = "quantization_scheme.input_activations"
        quant_args = getattr_chain(module, quant_args_attr, None)
        quant_enabled = getattr(module, "quantization_enabled", True)
        if quant_args is not None and quant_enabled:
            key_states = forward_quantize(module, key_states, "k", quant_args)
            value_states = forward_quantize(module, value_states, "v", quant_args)

        # original cache
        if self.past_key_values is not None:
            ret = self.past_key_values().update(
                key_states, value_states, *args, **kwargs
            )
        else:
            ret = (key_states, value_states)
        self.past_key_values = None

        return ret

    def add_past_key_values(self, past_key_values: Optional[Cache]):
        if past_key_values is not None:
            self.past_key_values = ref(past_key_values)
        else:
            self.past_key_values = None


# ----- initialize ----- #


def _kv_cache_attention_hook(
    module: Module, args: List[Any], kwargs: Dict[str, Any]
) -> Tuple[List[Any], Dict[str, Any]]:
    """
    Hook which should be called before each quantized attention forward pass.
    This hook dynamically replaces the `past_key_values` kwarg to the attention
    forward function.

    The original kvcache object is assigned to QuantizedKVCache().past_key_values
    as a weakref to maintain original cache functionality and compute savings
    """
    _past_kv_name = (
        "past_key_values"  # transformers#39956
        if "past_key_values" in inspect.signature(module.forward).parameters
        else "past_key_value"
    )
    past_key_values: Optional[Cache] = kwargs.get(_past_kv_name, None)

    cache: QuantizedKVCache = getattr(module, KV_CACHE_ATTR)
    cache.add_past_key_values(past_key_values)
    kwargs[_past_kv_name] = cache

    return args, kwargs


def initialize_hooked_kv_cache(model: PreTrainedModel, module: Module):
    """
    Initialize a `QuantizedKVCache` instance attached to attention

    :param model: parent model of attention module
    :param module: attention module to initialize with
    """
    if not hasattr(module, KV_CACHE_ATTR):
        module.register_module(KV_CACHE_ATTR, QuantizedKVCache(model.config, module))
        module.register_forward_pre_hook(_kv_cache_attention_hook, with_kwargs=True)


# ----- hooks ----- #


def register_key_hook(
    module: Module, hook: Callable[[Module, Tensor], Optional[Tensor]]
) -> RemovableHandle:
    """
    Register a hook which takes post-rope key states as an argument and
    returns the modified key states or `None`

    :param module: attention module to add hook to
    :param hook: key hook function
    """
    kv_cache: QuantizedKVCache = getattr(module, KV_CACHE_ATTR)

    def _hook(cache: QuantizedKVCache, args, kwargs):
        bound = inspect.signature(cache.forward).bind(*args, **kwargs)
        value = hook(module, bound.arguments["key_states"])
        if value is not None:
            bound.arguments["key_states"] = value

        return bound.args, bound.kwargs

    return kv_cache.register_forward_pre_hook(_hook, with_kwargs=True)


def register_value_hook(
    module: Module, hook: Callable[[Module, Tensor], Optional[Tensor]]
) -> RemovableHandle:
    """
    Register a hook which takes value states as an argument and
    returns the modified value states or `None`

    :param module: attention module to add hook to
    :param hook: value hook function
    """
    kv_cache: QuantizedKVCache = getattr(module, KV_CACHE_ATTR)

    def _hook(cache: QuantizedKVCache, args, kwargs):
        bound = inspect.signature(cache.forward).bind(*args, **kwargs)
        value = hook(module, bound.arguments["value_states"])
        if value is not None:
            bound.arguments["value_states"] = value

        return bound.args, bound.kwargs

    return kv_cache.register_forward_pre_hook(_hook, with_kwargs=True)



# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import shutil
from typing import Optional
from unittest.mock import MagicMock

import pytest
import torch
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    DEFAULT_QUANTIZATION_METHOD,
    FP8_E4M3_DATA,
    QuantizationArgs,
    QuantizationConfig,
    QuantizationScheme,
    QuantizationStatus,
    QuantizationStrategy,
    QuantizationType,
)
from compressed_tensors.quantization.lifecycle import apply_quantization_config
from compressed_tensors.utils import is_match, match_named_modules
from tests.testing_utils import requires_accelerate
from transformers import AutoModelForCausalLM


@pytest.fixture(scope="module", autouse=True)
def cleanup_model_cache():
    """Clean up the test model cache directory after all tests complete."""
    yield
    try:
        shutil.rmtree("test-apply-model-cache", ignore_errors=True)
    except Exception:
        pass


@pytest.fixture
def mock_model():
    model = MagicMock()
    model.named_modules.return_value = [
        ("layer1", MagicMock()),
        ("layer2", MagicMock()),
        ("layer3", MagicMock()),
    ]
    return model


@pytest.fixture
def mock_module():
    return MagicMock()


@pytest.fixture
def llama_stories_model():
    return AutoModelForCausalLM.from_pretrained(
        "Xenova/llama2.c-stories15M",
        torch_dtype="auto",
        cache_dir="test-apply-model-cache",
    )


def test_target_prioritization(mock_frozen):
    # tests that the config_groups are applied in the correct order
    # of priority, where exact layer name > regex > module name
    config = {
        "quant_method": "compressed-tensors",
        "format": "fakequant",
        "config_groups": {
            "group_1": {
                "weights": {
                    "num_bits": 8,
                },
                "targets": ["Linear"],
            },
            "group_2": {
                "weights": {
                    "num_bits": 4,
                },
                "targets": ["re:.*down_proj"],
            },
            "group_3": {
                "weights": {
                    "num_bits": 2,
                },
                "targets": ["model.layers.0.mlp.down_proj"],
            },
        },
    }

    model = AutoModelForCausalLM.from_pretrained(
        "HuggingFaceM4/tiny-random-LlamaForCausalLM",
        torch_dtype="auto",
        cache_dir="test-apply-model-cache",
    )
    model.eval()

    config = QuantizationConfig(**config)
    config.quantization_status = QuantizationStatus.CALIBRATION
    apply_quantization_config(model, config)
    mock_frozen(model)

    for name, module in model.named_modules():
        if name == "model.layers.0.mlp.down_proj":
            assert module.quantization_scheme.weights.num_bits == 2
        elif re.match(".*down_proj", name):
            assert module.quantization_scheme.weights.num_bits == 4
        elif isinstance(module, torch.nn.Linear):
            assert module.quantization_scheme.weights.num_bits == 8


def test_apply_quantization_config_tinyllama():
    quant_config = get_sample_tinyllama_quant_config(
        status=QuantizationStatus.INITIALIZED
    )
    model = get_tinyllama_model()

    # check that model is not already quantized
    for module in model.modules():
        _test_layer_quantization_status(module, inputs=False, weights=False)

    # apply quant config to model
    apply_quantization_config(model, quant_config)

    # check for correct application of quant config
    for quant_scheme in quant_config.config_groups.values():
        for name, module in match_named_modules(
            model, quant_scheme.targets, quant_config.ignore
        ):
            _test_layer_quantization_status(
                module,
                inputs=quant_scheme.input_activations is not None,
                weights=quant_scheme.weights is not None,
                expected_status=QuantizationStatus.INITIALIZED,
            )


@pytest.mark.parametrize(
    "config",
    [
        QuantizationConfig(
            config_groups={
                "linear": QuantizationScheme(
                    targets=["Linear"],
                    input_activations=QuantizationArgs(
                        num_bits=8,
                        type="float",
                        strategy="tensor",
                        scale_dtype=FP8_E4M3_DATA.dtype,
                        zp_dtype=torch.float,
                    ),
                )
            }
        ),
        QuantizationConfig(
            config_groups={
                "linear": QuantizationScheme(
                    targets=["Linear"],
                    input_activations=QuantizationArgs(
                        num_bits=8,
                        type="float",
                        strategy="tensor",
                        scale_dtype=FP8_E4M3_DATA.dtype,
                        zp_dtype=torch.float,
                    ),
                )
            },
            ignore=[
                "model.layers.0.self_attn.q_proj",
                "model.layers.1.self_attn.k_proj",
                "model.layers.2.self_attn.v_proj",
            ],
        ),
        QuantizationConfig(
            config_groups={},
            kv_cache_scheme=QuantizationArgs(
                num_bits=8,
                type="float",
                strategy="tensor",
                scale_dtype=FP8_E4M3_DATA.dtype,
                zp_dtype=torch.float,
            ),
        ),
        QuantizationConfig(
            config_groups={
                "attention": QuantizationScheme(
                    targets=["LlamaAttention"],
                    input_activations=QuantizationArgs(
                        num_bits=8,
                        type="float",
                        strategy="tensor",
                        scale_dtype=FP8_E4M3_DATA.dtype,
                        zp_dtype=torch.float,
                    ),
                )
            },
            kv_cache_scheme=QuantizationArgs(
                num_bits=8,
                type="float",
                strategy="tensor",
                scale_dtype=FP8_E4M3_DATA.dtype,
                zp_dtype=torch.float,
            ),
        ),
    ],
)
def test_from_pretrained(config: QuantizationConfig):
    model = AutoModelForCausalLM.from_pretrained("nm-testing/llama2.c-stories15M")
    apply_quantization_config(model, config)
    _config = QuantizationConfig.from_pretrained(model)
    assert list(_config.config_groups.values()) == list(config.config_groups.values())
    assert _config.kv_cache_scheme == config.kv_cache_scheme
    assert _config.ignore == config.ignore


def test_serialize_config_tinyllama():
    quant_config = get_sample_tinyllama_quant_config()
    model = get_tinyllama_model()

    # check that model is not already quantized
    for module in model.modules():
        _test_layer_quantization_status(module, inputs=False, weights=False)

    # apply quant config to model
    apply_quantization_config(model, quant_config)

    serialized_config = QuantizationConfig.from_pretrained(model)
    assert len(serialized_config.config_groups) == 2
    assert serialized_config.config_groups["group_0"].targets == ["Embedding"]
    assert serialized_config.config_groups["group_0"].input_activations is None
    assert serialized_config.config_groups["group_1"].targets == ["Linear"]
    assert serialized_config.config_groups["group_1"].input_activations is not None
    assert serialized_config.format == CompressionFormat.dense.value
    assert serialized_config.quant_method == DEFAULT_QUANTIZATION_METHOD
    assert serialized_config.ignore == ["model.layers.1.mlp.down_proj"]
    if serialized_config.global_compression_ratio is not None:
        assert serialized_config.global_compression_ratio > 1.0
        assert serialized_config.global_compression_ratio < 8.0


def _test_layer_quantization_status(
    module,
    inputs: bool,
    weights: bool,
    expected_status: Optional[QuantizationStatus] = None,
    expected_dtype: Optional[torch.dtype] = None,
):
    # check if quantization is applied at all (true if inputs or weights targeted)
    quantized = inputs or weights
    assert hasattr(module, "quantization_scheme") == quantized
    assert hasattr(module, "quantization_status") == quantized
    if expected_status is not None:
        assert module.quantization_status is expected_status

    # check inputs matches expected
    assert hasattr(module, "input_scale") == inputs
    assert hasattr(module, "input_zero_point") == inputs

    # check weights matches expected
    assert hasattr(module, "weight_scale") == weights
    assert hasattr(module, "weight_zero_point") == weights
    if weights and expected_dtype is not None:
        assert module.weight.dtype is expected_dtype


def get_tinyllama_model():
    return AutoModelForCausalLM.from_pretrained(
        "TinyLlama/TinyLlama-1.1B-intermediate-step-1431k-3T",
        torch_dtype="auto",
        cache_dir="test-apply-model-cache",
    )


def get_sample_tinyllama_quant_config(
    status: QuantizationStatus = QuantizationStatus.FROZEN,
):
    config_dict = {
        "quant_method": "compressed-tensors",
        "format": "fakequant",
        "quantization_status": status,
        "global_compression_ratio": None,
        "config_groups": {
            "group_1": {
                "weights": {
                    "num_bits": 8,
                    "type": "int",
                    "symmetric": True,
                    "strategy": "tensor",
                },
                "input_activations": {
                    "num_bits": 8,
                    "type": "int",
                    "symmetric": True,
                    "strategy": "tensor",
                },
                "targets": ["Linear"],
            },
            "group_2": {
                "weights": {
                    "num_bits": 8,
                    "type": "int",
                    "symmetric": False,
                    "strategy": "tensor",
                },
                "input_activations": None,
                "targets": ["Embedding"],
            },
        },
        "ignore": ["LlamaRotaryEmbedding", "model.layers.1.mlp.down_proj"],
    }
    return QuantizationConfig.model_validate(config_dict)


@requires_accelerate()
@pytest.mark.parametrize(
    "target,should_raise_warning",
    [
        [("Linear",), False],
        [("Linear", "re:.*foobarbaz"), True],
    ],
)
def test_apply_quantization_config(caplog, target, should_raise_warning):
    import logging

    # load a dense, unquantized tiny llama model
    model = get_tinyllama_model()
    quantization_config_dict = {
        "quant_method": "compressed-tensors",
        "format": "pack-quantized",
        "global_compression_ratio": None,
        "config_groups": {
            "group_1": {
                "weights": {
                    "num_bits": 4,
                    "type": "int",
                    "symmetric": False,
                    "strategy": "tensor",
                },
                "targets": target,
            }
        },
        "ignore": ["lm_head", "re:.*gate"],
    }

    config = QuantizationConfig(**quantization_config_dict)
    config.quantization_status = QuantizationStatus.CALIBRATION

    # mismatch in the ignore key of quantization_config_dict
    with caplog.at_level(logging.WARNING):
        apply_quantization_config(model, config)
        if should_raise_warning:
            assert len(caplog.text) > 0
        else:
            assert len(caplog.text) == 0


def test_multi_apply_quantization_config():
    """
    Ensure that multiple quantization configs are applied correctly
    If quantization config was previously applied to a module,
    those changes should be reset for newly applied quantization config
    """
    model = get_tinyllama_model()

    # FP8 applied to self_attn
    qconfig1 = QuantizationConfig(
        config_groups={
            "group_0": QuantizationScheme(
                targets=[
                    r"re:.*self_attn\.(k|q|o|v)_proj$",
                ],
                weights=QuantizationArgs(
                    num_bits=8,
                    type=QuantizationType.FLOAT,
                    strategy=QuantizationStrategy.TENSOR,
                    symmetric=True,
                    dynamic=False,
                ),
                input_activations=QuantizationArgs(
                    num_bits=8,
                    type=QuantizationType.FLOAT,
                    strategy=QuantizationStrategy.TENSOR,
                    symmetric=True,
                    dynamic=False,
                ),
            )
        },
        ignore=["lm_head"],
    )
    # W4A16_ASYM applied to mlp and self_attn.o_proj to validate overwriting
    qconfig2 = QuantizationConfig(
        config_groups={
            "group_0": QuantizationScheme(
                targets=[
                    r"re:.*mlp\.(down|gate|up)_proj$",
                    r"re:.*self_attn\.o_proj$",
                ],
                weights=QuantizationArgs(
                    num_bits=4,
                    type=QuantizationType.INT,
                    strategy=QuantizationStrategy.GROUP,
                    group_size=128,
                    symmetric=False,
                    dynamic=False,
                ),
            )
        },
        ignore=["lm_head"],
    )

    apply_quantization_config(model, qconfig1)
    apply_quantization_config(model, qconfig2)
    for name, module in model.named_modules():
        if is_match(
            name, module, qconfig2.config_groups["group_0"].targets, qconfig2.ignore
        ):
            # assert W4A16_ASYM parameters are present with correct shape
            # and FP8 parameters have been removed
            assert not hasattr(module, "input_scale")
            assert not hasattr(module, "input_zero_point")
            weight_scale = getattr(module, "weight_scale", None)
            assert (
                weight_scale is not None
                and weight_scale.shape[:-1] == module.weight.shape[:-1]
                and weight_scale.shape[-1] == module.weight.shape[-1] / 128
            )
            weight_zero_point = getattr(module, "weight_zero_point", None)
            assert (
                weight_zero_point is not None
                and weight_zero_point.shape[:-1] == module.weight.shape[:-1]
                and weight_zero_point.shape[-1] == module.weight.shape[-1] / 128
            )

        elif is_match(
            name, module, qconfig1.config_groups["group_0"].targets, qconfig1.ignore
        ):
            # assert FP8 scheme parameters are present with correct shape
            input_scale = getattr(module, "input_scale", None)
            assert input_scale is not None and input_scale.shape == torch.Size([1])
            input_zero_point = getattr(module, "input_zero_point", None)
            assert (
                input_zero_point is not None
                and input_zero_point.shape == torch.Size([1])
            )
            weight_scale = getattr(module, "weight_scale", None)
            assert weight_scale is not None and weight_scale.shape == torch.Size([1])
            weight_zero_point = getattr(module, "weight_zero_point", None)
            assert (
                weight_zero_point is not None
                and weight_zero_point.shape == torch.Size([1])
            )


@requires_accelerate()
def test_apply_kv_cache():
    from accelerate import init_empty_weights

    with init_empty_weights():
        model = AutoModelForCausalLM.from_pretrained("nm-testing/llama2.c-stories15M")

    args = QuantizationArgs(
        num_bits=8,
        type="float",
        strategy="tensor",
        scale_dtype=FP8_E4M3_DATA.dtype,
        zp_dtype=torch.float,
    )
    config = QuantizationConfig(config_groups={}, kv_cache_scheme=args)

    apply_quantization_config(model, config)

    for layer in model.model.layers:
        assert getattr(layer.self_attn, "quantization_scheme").input_activations == args
        assert hasattr(layer.self_attn, "k_scale")
        assert hasattr(layer.self_attn, "v_scale")


@requires_accelerate()
def test_apply_attention():
    from accelerate import init_empty_weights

    with init_empty_weights():
        model = AutoModelForCausalLM.from_pretrained("nm-testing/llama2.c-stories15M")

    scheme = QuantizationScheme(
        targets=["LlamaAttention"],
        input_activations=QuantizationArgs(
            num_bits=8,
            type="float",
            strategy="tensor",
            scale_dtype=FP8_E4M3_DATA.dtype,
            zp_dtype=torch.float,
        ),
    )
    config = QuantizationConfig(config_groups={"attention": scheme})

    apply_quantization_config(model, config)

    for layer in model.model.layers:
        assert getattr(layer.self_attn, "quantization_scheme") == scheme
        assert hasattr(layer.self_attn, "q_scale")
        assert hasattr(layer.self_attn, "k_scale")
        assert hasattr(layer.self_attn, "v_scale")
